from .bsts import ObservationModelManager
import patsy
import numpy as np
import BayesBoom.boom as boom
import BayesBoom.spikeslab as spikeslab
import scipy
from numbers import Number


class StateSpaceLogitModelFactory:
    def __init__(self, formula):
        """
        Args:
          formula: A string describing how the predictor variables are to be
            constructed from a data frame.  For example "y ~ x1 + x2 + x2*x3".
            If the formula is None then the model will contain no regression
            component.
        """
        if formula is not None and not isinstance(formula, str):
            raise Exception("formula must either be None or a string")
        self._formula = formula
        self.predictor_names = None

    def create_model(self, prior, data, **kwargs):
        if data is not None:
            response, predictors = patsy.dmatrices(self._formula, data)
            self.predictor_names = predictors.design_info.term_names
            extra_args = {**kwargs}
            trials = extra_args.get("trials", 1)
            if isinstance(trials, Number):
                trials = np.full(len(response), trials)
            observed = np.isfinite(response)
            self._model = boom.StateSpacePoissonModel(
                boom.Vector(response),
                boom.Vector(trials),
                boom.Matrix(predictors),
                observed)
        elif prior is not None:
            xdim = len(prior._prior_inclusion_probabilities)
            self._model = boom.StateSpaceLogitModel(xdim)
            response = None
            predictors = None
            trials = None
        else:
            raise Exception("At least one of 'data' or 'prior' is needed.")

        logit_reg = self._model.observation_model
        prior = self._verify_prior(prior, response, predictors, trials,
                                   **kwargs)
        self._prior = prior
        observation_model_sampler = prior.create_sampler(logit_reg, assign=True)

        sampler = boom.StateSpacePoissonPosteriorSampler(
            self._model, observation_model_sampler)
        self._model.set_method(sampler)
        self._original_series = response
        return self._model

    def create_observation_model_manager(self):
        return LogitObservationModelManager(
            xdim=self._model.observation_model.xdim,
            formula=self._formula)

    def _verify_prior(self, prior, response, predictors, trials, **kwargs):
        if prior is None:
            prior = spikeslab.LogitZellnerPrior(
                predictors=predictors,
                response=response,
                trials=trials,
                **kwargs)
        if not isinstance(prior, spikeslab.LogitZellnerPrior):
            raise Exception(
                "Expected 'prior' to be a 'spikeslab.LogitZellnerPrior'."
            )
        return prior


class LogitObservationModelManager(ObservationModelManager):
    def __init__(self, xdim: int, formula: str):
        #  TODO
        self._xdim = xdim
        self._formula = formula

    @property
    def has_regression(self):
        return self._xdim > 1 and self._formula is not None

    @property
    def niter(self):
        if hasattr(self, "_coefficients"):
            return self._coefficients.shape[0]
        else:
            return 0

    def allocate_space(self, niter: int, time_dimension: int):
        self._coefficients = scipy.sparse.lil_matrix((niter, self._xdim))
        if self.has_regression:
            self._regression_contribution = np.empty((niter, time_dimension))

    def record_draw(self, iteration: int, model):
        self._coefficients[iteration, :] = spikeslab.sparsify(
            model.observation_model.coef)
        if self.has_regression:
            self._regression_contribution[iteration, :] = (
                model.regression_contribution.to_numpy()
            )

    def restore_draw(self, iteration: int, model):
        spikeslab.set_glm_coefs(model.observation_model.coef,
                                self._coefficients[iteration, :])

    def format_prediction_data(self, prediction_data, **kwargs):
        if isinstance(prediction_data, int):
            formatted = {
                "forecast_horizon": prediction_data,
                "predictors": boom.Matrix(np.ones((int(prediction_data), 1)))
            }
        else:
            predictor_matrix = patsy.dmatrix(self._formula,
                                             data=prediction_data)
            xnames = predictor_matrix.design_info.term_names
            formatted = {
                "forecast_horizon": prediction_data.shape[0],
                "predictors": boom.Matrix(predictor_matrix),
                "xnames": xnames,
            }
        extra_args = {**kwargs}
        trials = extra_args.get("trials", 1)
        if isinstance(trials, Number):
            trials = np.full(formatted["forecast_horizon"], trials)
        else:
            trials = np.array(trials)
        formatted["trials"] = boom.Vector(trials)

        return formatted

    def predict(self, model, formatted_prediction_data, boom_final_state, rng,
                separate_components=False, **kwargs):
        if separate_components:
            draw = model.simulate_forecast_components(
                rng,
                formatted_prediction_data["predictors"],
                formatted_prediction_data["trials"],
                boom_final_state)
        else:
            draw = model.simulate_forecast(
                rng,
                formatted_prediction_data["predictors"],
                formatted_prediction_data["trials"],
                boom_final_state)
        return draw.to_numpy()



import numpy as np
import pandas as pd


def simulate_data(sample_size: int, numeric_dim=1, cat_levels={}):
    numerics = np.random.randn(sample_size, numeric_dim)
    numeric_names = ["X" + str(i+1) for i in range(numeric_dim)]
    data = pd.DataFrame(numerics, columns=numeric_names)
    for vname, levels in cat_levels.items():
        data[vname] = np.random.choice(levels, size=sample_size)
    return data



import BayesBoom.impute as impute
import BayesBoom.test_utils as tu

import unittest
import numpy as np


class ImputerTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        self.data = tu.simulate_data(
            sample_size=100, numeric_dim=3, cat_levels={
                "colors": ["red", "blue", "green"],
                "states": ["CA", "TX", "NY", "OR", "MA"],
            })
        self.model = impute.MixedDataImputer()

    def test_empty_model(self):
        self.assertEqual(self.model.nclusters, 0)
        self.assertEqual(self.model.niter, 0)

        self.model.find_atoms(self.data)
        self.assertEqual(self.model._numeric_colnames, ['X1', 'X2', 'X3'])
        self.assertTrue(isinstance(self.model._atoms, dict))
        self.assertEqual(self.model._atoms["X1"], [])

    def test_fit_imputation_model(self):
        data = self.data
        data["X1"][:10] = np.NaN
        data["X2"][5:15] = np.NaN
        self.model.find_atoms(data)
        num_clusters = 5
        niter = 100
        self.model.train_model(data=data, niter=niter,
                               nclusters=num_clusters)
        self.assertEqual(num_clusters, self.model.nclusters)
        num_states = 5
        num_colors = 3
        xdim = 1 + (num_colors - 1) + (num_states - 1)
        self.assertEqual(self.model.xdim, xdim)
        self.assertEqual(3, self.model.ydim)
        self.assertEqual(self.model.atom_probs["X1"].shape,
                         (niter, num_clusters, 1))


_debug_mode = True

if _debug_mode:
    import pdb
    rig = ImputerTest()
    rig.setUp()
    rig.test_fit_imputation_model()

elif __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import pandas as pd
import BayesBoom.boom as boom
import BayesBoom.dynreg as dynreg
import BayesBoom.spikeslab as ss
import BayesBoom.R as R
import matplotlib.pyplot as plt

import pdb


class TestDynamicRegression(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    @staticmethod
    def simulate_data_from_model(time_dimension: int,
                                 typical_sample_size: int,
                                 xdim: int,
                                 residual_sd: float,
                                 unscaled_innovation_sd: np.ndarray,
                                 p00: np.ndarray,
                                 p11: np.ndarray):
        from BayesBoom.R import rmarkov
        inclusion = np.full((xdim, time_dimension), -1)
        p00 = p00.ravel()
        p11 = p11.ravel()
        for j in range(xdim):
            P = np.array(
                [
                    [p00[j], 1 - p00[j]],
                    [1 - p11[j], p11[j]]
                ]
            )
            inclusion[j, :] = rmarkov(time_dimension, P)

        coefficients = np.zeros((xdim, time_dimension))
        for j in range(xdim):
            sd = unscaled_innovation_sd[j] * residual_sd
            for t in range(time_dimension):
                prev = 0 if t == 0 else coefficients[j, t-1]
                coefficients[j, t] = inclusion[j, t] * (
                    prev + np.random.randn(1) * sd
                )

        data = []
        xnames = ["X" + str(i) for i in range(xdim)]
        xnames[0] = "(Intercept)"
        for t in range(time_dimension):
            sample_size = np.random.poisson(typical_sample_size, 1)[0]
            X = np.random.randn(sample_size, xdim)
            X[:, 0] = 1.0
            yhat = X @ coefficients[:, t]
            y = yhat + residual_sd * np.random.randn(sample_size)
            chunk = pd.DataFrame(X, columns=xnames)
            chunk["y"] = y
            chunk["timestamp"] = t
            data.append(chunk)

        full_data = pd.concat(data, axis=0, ignore_index=True)
        return full_data, coefficients, inclusion

    def test_mcmc(self):
        xdim = 4
        true_residual_sd = .25
        data, coefficients, inclusion = self.simulate_data_from_model(
            time_dimension=200,
            typical_sample_size=5000,
            xdim=xdim,
            residual_sd=true_residual_sd,
            unscaled_innovation_sd=np.array([.01] * xdim),
            p00=np.array([.95] * xdim),
            p11=np.array([.99] * xdim),
        )
        model = dynreg.SparseDynamicRegressionModel(
            "y ~ " + ss.dot(data, ["y", "timestamp", "(Intercept)"]),
            data=data,
            timestamps="timestamp",
            niter=100,
            residual_precision_prior=R.SdPrior(true_residual_sd, 1),
            seed=8675309)

        model.plot()
        for i in range(4):
            self.assertEqual("", boom.check_stochastic_process(
                boom.Matrix(model._beta_draws[:, i, :]),
                boom.Vector(coefficients[i, :]),
                confidence=.95,
                sd_ratio_threshold=10000,  # Turn off the sd_ratio check.
            ))

        posterior_mean_residual_sd = np.mean(model._residual_sd_draws[10:])
        self.assertGreater(posterior_mean_residual_sd, true_residual_sd - .02)
        self.assertLess(posterior_mean_residual_sd, true_residual_sd + .02)

        sd_fig, sd_ax = plt.subplots(1, 2)
        model.plot_residual_sd(ax=sd_ax[0])
        model.plot_residual_sd(ax=sd_ax[1], type="ts")
        # sd_fig.show()

        size_fig, size_ax = plt.subplots(1, 1)
        model.plot_size(ax=size_ax)
        # size_fig.show()

        # fig, ax = plt.subplots(2, 2)
        # which = 0
        # for i in range(2):
        #     for j in range(2):
        #         R.plot_dynamic_distribution(model._beta_draws[:, which, :],
        #                                     ax=ax[i, j])
        #         R.lines(np.arange(coefficients.shape[1]),
        #                 coefficients[which, :], color="blue", ax=ax[i, j])
        #         which += 1
        # fig.show()


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
#    import warnings
#    warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestDynamicRegression()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_mcmc()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import BayesBoom.boom as boom
import numpy as np
from .state_models import StateModel
import BayesBoom.R as R


class SemilocalLinearTrendStateModel(StateModel):
    """
    """

    def __init__(self,
                 y=None,
                 level_sigma_prior: R.SdPrior = None,
                 slope_mean_prior: R.NormalPrior = None,
                 slope_ar1_prior: R.Ar1CoefficientPrior = None,
                 slope_sigma_prior: R.SdPrior = None,
                 initial_level_prior: R.NormalPrior = None,
                 initial_slope_prior: R.NormalPrior = None,
                 sdy: float = None,
                 initial_y: float = None):
        """
        Args:
          y:  A numeric vector.  The time series to be modeled.
          level_sigma_prior: The prior distribution for the standard deviation
            of the increments in the level component of state.
          slope.mean_prior: The prior distribution for the mean of the AR1
            process for the slope component of state.
          slope_ar1_prior: The prior distribution for the ar1 coefficient in
            the slope component of state.
          slope_sigma_prior: The prior distribution for the standard deviation
            of the increments in the slope component of state.
          initial_level_prior: The prior distribution for the level component
            of state at the time of the first observation.
          initial_slope_prior: The prior distribution for the slope component
            of state at the time of the first observation.
          sdy: The standard deviation of y.  This can be ignored if y is
            provided, or if all the required prior distributions are supplied
            directly.
          initial.y: The initial value of y.  This can be omitted if y is
            provided.
        """

        # ----------------------------------------------------------------------
        # Validate all priors, generating default values as needed.
        if sdy is None:
            sdy = np.nanstd(y)
        if initial_y is None:
            initial_y = y[0]
        self.level_sigma_prior = self._validate_level_sigma_prior(
            level_sigma_prior, sdy)
        self.slope_mean_prior = self._validate_slope_mean_prior(
            slope_mean_prior, sdy)
        self.slope_ar1_prior = self._validate_slope_ar1_prior(
            slope_ar1_prior, sdy)
        self.slope_sigma_prior = self._validate_slope_sigma_prior(
            slope_sigma_prior, sdy)
        initial_level_prior = self._validate_initial_level_prior(
            initial_level_prior, initial_y, sdy)
        initial_slope_prior = self._validate_initial_slope_prior(
            initial_slope_prior, sdy)

        self._create_model()
        self._set_initial_state_distribution(
            initial_level_prior, initial_slope_prior)

    @property
    def label(self):
        return "Trend (semilocal)"

    def __repr__(self):
        return (
            "A SemilocalLinearTrendStateModel with \n"
            f"level_sd = {self._state_model.level_sd} \n"
            f"slope_sd = {self._state_model.slope_sd}\n"
            f"slope_ar_coefficient = {self._state_model.slope_ar_coefficient}\n"
            f"slope_mean = {self._state_model.slope_mean}\n"
            )

    @property
    def state_dimension(self):
        return self._state_model.state_dimension

    @property
    def state_error_dimension(self):
        return self._state_model.state_error_dimension

    # TODO: change the names once they're defined in the pb11 module.
    def allocate_space(self, niter, time_dimension):
        self.level_sigma = np.empty(niter)
        self.slope_sigma = np.empty(niter)
        self.slope_ar1 = np.empty(niter)
        self.slope_mean = np.empty(niter)
        self._state_contribution = np.empty((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self.level_sigma[iteration] = self._state_model.level_sd
        self.slope_sigma[iteration] = self._state_model.slope_sd
        self.slope_ar1[iteration] = self._state_model.slope_ar_coefficient
        self.slope_mean[iteration] = self._state_model.slope_mean
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

    def restore_state(self, iteration):
        self._state_model.set_level_sd(self.level_sigma[iteration])
        self._state_model.set_slope_sd(self.slope_sigma[iteration])
        self._state_model.set_slope_ar_coefficient(self.slope_ar1[iteration])
        self._state_model.set_slope_mean(self.slope_mean[iteration])

    @property
    def state_contribution(self):
        return self._state_contribution

    @staticmethod
    def _validate_level_sigma_prior(level_sigma_prior, sdy):
        if level_sigma_prior is None:
            level_sigma_prior = R.SdPrior(.01 * sdy, upper_limit=sdy)
        if not isinstance(level_sigma_prior, R.SdPrior):
            raise Exception("Wrong type passed for level_sigma_prior.  "
                            "Expected an R.SdPrior")
        return level_sigma_prior

    @staticmethod
    def _validate_slope_sigma_prior(slope_sigma_prior, sdy):
        if slope_sigma_prior is None:
            slope_sigma_prior = R.SdPrior(.01 * sdy, upper_limit=sdy)
        if not isinstance(slope_sigma_prior, R.SdPrior):
            raise Exception("Wrong type passed for slope_sigma_prior.  "
                            "Expected an R.SdPrior")
        return slope_sigma_prior

    @staticmethod
    def _validate_slope_ar1_prior(slope_ar1_prior, sdy):
        if slope_ar1_prior is None:
            slope_ar1_prior = R.Ar1CoefficientPrior()
        if not isinstance(slope_ar1_prior, R.Ar1CoefficientPrior):
            raise Exception("Wrong type passed for slope_ar1_prior.  "
                            "Expected an R.Ar1CoefficientPrior")
        return slope_ar1_prior

    @staticmethod
    def _validate_slope_mean_prior(slope_mean_prior, sdy):
        if slope_mean_prior is None:
            slope_mean_prior = R.NormalPrior(0, sdy)
        if not isinstance(slope_mean_prior, R.NormalPrior):
            raise Exception("Wrong type passed for slope_mean_prior.  "
                            "Expected an R.NormalPrior")
        return slope_mean_prior

    @staticmethod
    def _validate_initial_slope_prior(initial_slope_prior, sdy):
        if initial_slope_prior is None:
            initial_slope_prior = R.NormalPrior(0, sdy)
        if not isinstance(initial_slope_prior, R.NormalPrior):
            raise Exception("Wrong type for initial_slope_prior.  "
                            "Expected an R.NormalPrior.")
        return initial_slope_prior

    @staticmethod
    def _validate_initial_level_prior(initial_level_prior, initial_y, sdy):
        if initial_level_prior is None:
            initial_level_prior = R.NormalPrior(initial_y, sdy)
        if not isinstance(initial_level_prior, R.NormalPrior):
            raise Exception("Wrong type for initial_level_prior.  "
                            "Expected an R.NormalPrior.")
        return initial_level_prior

    def __getstate__(self):
        payload = {
            "state_index": self._state_index,
            "initial_state_mean":
            self._state_model.initial_state_mean.to_numpy(),
            "initial_state_variance":
            self._state_model.initial_state_variance.to_numpy(),
            "level_sigma_prior": self.level_sigma_prior.__getstate__(),
            "slope_sigma_prior": self.slope_sigma_prior.__getstate__(),
            "slope_ar1_prior": self.slope_ar1_prior.__getstate__(),
            "slope_mean_prior": self.slope_mean_prior.__getstate__(),
        }

        if hasattr(self, "level_sigma"):
            payload["level_sigma"] = self.level_sigma
        if hasattr(self, "slope_sigma"):
            payload["slope_sigma"] = self.slope_sigma
        if hasattr(self, "slope_ar1"):
            payload["slope_ar1"] = self.slope_ar1
        if hasattr(self, "slope_mean"):
            payload["slope_mean"] = self.slope_mean
        if hasattr(self, "_state_contribution"):
            payload["state_contribution"] = self._state_contribution

        return payload

    def __setstate__(self, payload):
        self.level_sigma_prior = R.SdPrior(1, 1)
        self.slope_sigma_prior = R.SdPrior(1, 1)
        self.slope_ar1_prior = R.Ar1CoefficientPrior()
        self.slope_mean_prior = R.NormalPrior(0, 1)
        self.level_sigma_prior.__setstate__(payload["level_sigma_prior"])
        self.slope_sigma_prior.__setstate__(payload["slope_sigma_prior"])
        self.slope_ar1_prior.__setstate__(payload["slope_ar1_prior"])
        self.slope_mean_prior.__setstate__(payload["slope_mean_prior"])
        self._create_model()

        initial_state_mean = payload["initial_state_mean"]
        initial_state_variance = payload["initial_state_variance"]
        self._set_initial_state_distribution(
            initial_level_prior=R.NormalPrior(
                initial_state_mean[0], np.sqrt(initial_state_variance[0, 0])),
            initial_slope_prior=R.NormalPrior(
                initial_state_mean[0], np.sqrt(initial_state_variance[0, 0]))
        )

        self.level_sigma = payload.get("level_sigma", None)
        self.slope_sigma = payload.get("slope_sigma", None)
        self.slope_ar1 = payload.get("slope_ar1", None)
        self.slope_mean = payload.get("slope_mean", None)
        self._state_contribution = payload.get("state_contribution", None)

    def _create_model(self):
        """
        Creaate self._state_model and assign it a posterior sampler.

        The four priors (level_sigma, slope_sigma, slope_mean, and slope_ar1)
        must have been created and saved to self.
        """
        level_model = boom.ZeroMeanGaussianModel(
            self.level_sigma_prior.initial_value)
        level_sampler = boom.ZeroMeanGaussianConjSampler(
            level_model,
            self.level_sigma_prior.boom(),
            boom.GlobalRng.rng)
        level_sampler.set_sigma_upper_limit(self.level_sigma_prior.upper_limit)
        level_model.set_method(level_sampler)

        slope_model = boom.NonzeroMeanAr1Model(
            self.slope_mean_prior.initial_value,
            self.slope_ar1_prior.initial_value,
            self.slope_sigma_prior.initial_value)

        slope_sampler = boom.NonzeroMeanAr1Sampler(
            slope_model,
            self.slope_mean_prior.boom(),
            self.slope_ar1_prior.boom(),
            self.slope_sigma_prior.boom())

        slope_sampler.set_sigma_upper_limit(self.slope_sigma_prior.upper_limit)
        if self.slope_ar1_prior.force_stationary:
            slope_sampler.force_stationary()
        if self.slope_ar1_prior.force_positive:
            slope_sampler.force_ar1_positive()

        # ----------------------------------------------------------------------
        # Create the state model.
        self._state_model = boom.SemilocalLinearTrendStateModel(
            level_model, slope_model)

        # BOOM models can have more than one posterior sampler if the samplers
        # focus on different parts of the job.
        self._state_model.set_method(level_sampler)
        self._state_model.set_method(slope_sampler)

    def _set_initial_state_distribution(
            self, initial_level_prior, initial_slope_prior):
        """
        Set the initial distribution of the state at time 0.
        """
        self._state_model.set_initial_level_mean(
            initial_level_prior.mu)
        self._state_model.set_initial_level_sd(
            initial_level_prior.sigma)
        self._state_model.set_initial_slope_mean(
            initial_slope_prior.mu)
        self._state_model.set_initial_slope_sd(
            initial_slope_prior.sigma)

        self._state_contribution = None



import BayesBoom.boom as boom
import numpy as np
from .state_models import StateModel
import BayesBoom.R as R


class StudentLocalLinearTrendStateModel(StateModel):
    """
    A local linear trend state model with Student T errors for both the level
    and slope components.  The model is

             y[t] = mu[t] + error[t]
          mu[t+1] = mu[t] + delta[t] + innovation0[t]
       delta[t+1] = delta[t] + innovation1[t]

    where innovation0[t] ~ Student(0, sigma0, nu0) and innovation1 ~ Student(0,
    sigma1, nu1).  The notation Y ~ Student(mu, sigma, nu) means that Y is mu +
    sigma * Z where Z follows a standard T distribution with nu "degrees of
    freedom".

    """

    def __init__(self,
                 y=None,
                 save_weights: bool = False,
                 level_sigma_prior: R.SdPrior = None,
                 level_nu_prior: R.DoubleModel = None,
                 slope_sigma_prior: R.SdPrior = None,
                 slope_nu_prior: R.DoubleModel = None,
                 initial_level_prior: R.NormalPrior = None,
                 initial_slope_prior: R.NormalPrior = None,
                 sdy: float = None,
                 initial_y: float = None):
        """
        Args:
          y:  The time series to be modeled.
          save_weights: If True then the mixing weights required to express the
            T distribution as a normal mixture are saved.  The weights are
            latent variables stored as a Monte Carlo ensemble.
          level_sigma_prior: Prior distribution on the standard deviation of
            the "level" innovations.
          level_nu_prior: Prior distribution for the tail thickness parameter
            'nu' for the level portion of the model.  The default is a U(.1,
            100) distribution.
          slope_sigma_prior: Prior distribution on the standard deviation of
            the "slope" innovations.
          level_nu_prior: Prior distribution for the tail thickness parameter
            'nu' for the slope portion of the model.  The default is a U(.1,
            100) distribution.
          initial_level_prior: Prior distribution on the value of the level at
            time 0.
          initial_slope_prior: Prior distribution on the value of the slope at
            time 0.
          sdy:  The standard deviation of the time series being modeled.
          initial_y: The value of y at time 0.

        Details:
          There are three ways to build this object.
          (1) Pass all four prior objects.
          (2) Pass zero or more prior objects, sdy, and initial_y.
          (3) Pass y.

          Option 3 is the most convenient.  Option 1 offers the most control.
        """
        self._save_weights = save_weights
        self._validate_priors(level_sigma_prior, level_nu_prior,
                              slope_sigma_prior, slope_nu_prior,
                              y, sdy)
        self._validate_initial_distributions(
            initial_level_prior, initial_slope_prior, y, sdy, initial_y)
        self._state_contribution = None
        self._build_state_model()

    def __repr__(self):
        return (
            "A StudentLocalLinearTrendStateModel"
        )

    @property
    def label(self):
        return "Trend (student)"

    @property
    def state_dimension(self):
        return 2

    @property
    def state_error_dimension(self):
        return 2

    @property
    def state_contribution(self):
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self.sigma_level = np.empty(niter)
        self.nu_level = np.empty(niter)
        self.sigma_slope = np.empty(niter)
        self.nu_slope = np.empty(niter)
        self._state_contribution = np.empty((niter, time_dimension))

        if self._save_weights:
            self._level_weights = np.empty((niter, time_dimension))
            self._slope_weights = np.empty((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self.sigma_level[iteration] = self._state_model.sigma_level
        self.nu_level[iteration] = self._state_model.nu_level
        self.sigma_slope[iteration] = self._state_model.sigma_slope
        self.nu_slope[iteration] = self._state_model.nu_slope
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

        if self._save_weights:
            self._level_weights[iteration, :] = (
                self._state_model.latent_level_weights.to_numpy()
            )
            self._slope_weights[iteration, :] = (
                self._state_model.latent_slope_weights.to_numpy()
            )

    def restore_state(self, iteration):
        self._state_model.set_sigma_level(self.sigma_level[iteration])
        self._state_model.set_nu_level(self.nu_level[iteration])
        self._state_model.set_sigma_slope(self.sigma_slope[iteration])
        self._state_model.set_nu_slope(self.nu_slope[iteration])

    def _validate_priors(self, level_sigma_prior, level_nu_prior,
                         slope_sigma_prior, slope_nu_prior,
                         y, sdy):
        if level_sigma_prior is None:
            sdy = self._compute_sdy(sdy, y, "level_sigma_prior")
            level_sigma_prior = R.SdPrior(
                sigma_guess=.01 * sdy,
                upper_limit=sdy)
        if not isinstance(level_sigma_prior, R.SdPrior):
            raise Exception("Unexpected type for level_sigma_prior.")

        if slope_sigma_prior is None:
            sdy = self._compute_sdy(sdy, y, "slope_sigma_prior")
            slope_sigma_prior = R.SdPrior(
                sigma_guess=.01 * sdy,
                upper_limit=sdy)
        if not isinstance(slope_sigma_prior, R.SdPrior):
            raise Exception("Unexpected type for slope_sigma_prior.")

        if level_nu_prior is None:
            level_nu_prior = R.UniformPrior(0.1, 100)
        if not isinstance(level_nu_prior, R.DoubleModel):
            raise Exception("Unexpected type for level_nu_prior.")

        if slope_nu_prior is None:
            slope_nu_prior = R.UniformPrior(0.1, 100)
        if not isinstance(slope_nu_prior, R.DoubleModel):
            raise Exception("Unexpected type for slope_nu_prior.")

        self._level_sigma_prior = level_sigma_prior
        self._slope_sigma_prior = slope_sigma_prior
        self._level_nu_prior = level_nu_prior
        self._slope_nu_prior = slope_nu_prior

    def _validate_initial_distributions(
            self, initial_level_prior, initial_slope_prior,
            y, sdy, initial_y):
        if initial_level_prior is None:
            sdy = self._compute_sdy(sdy, y, "initial_level_prior")
            if initial_y is None:
                if y is None:
                    raise Exception(
                        "One of initial_y, y, or initial_level_prior must be "
                        "specified.")
                else:
                    initial_y = y[0]
            initial_level_prior = R.NormalPrior(initial_y, sdy)
        if not isinstance(initial_level_prior, R.NormalPrior):
            raise Exception("Unexpected type for initial_level_prior.")
        self._initial_level_prior = initial_level_prior

        if initial_slope_prior is None:
            sdy = self._compute_sdy(sdy, y, "initial_slope_prior")
            initial_slope_prior = R.NormalPrior(0, sdy)
        if not isinstance(initial_slope_prior, R.NormalPrior):
            raise Exception("Unexpected type for initial_slope_prior.")
        self._initial_slope_prior = initial_slope_prior

    def _build_state_model(self):
        self._state_model = boom.StudentLocalLinearTrendStateModel()
        self._set_posterior_sampler()
        self._set_initial_distribution()
        self._state_model.set_nu_level(self._level_nu_prior.mean)
        self._state_model.set_nu_slope(self._slope_nu_prior.mean)



    def _set_posterior_sampler(self):
        """
        A utility called by the constructor.  See the __init__ method for
        argument documentation.
        """
        sampler = boom.StudentLocalLinearTrendPosteriorSampler(
            self._state_model,
            self._level_sigma_prior.boom(),
            self._level_nu_prior.boom(),
            self._slope_sigma_prior.boom(),
            self._slope_nu_prior.boom(),
            boom.GlobalRng.rng)
        if (
                self._level_sigma_prior.upper_limit > 0
                and np.isfinite(self._level_sigma_prior.upper_limit)
        ):
            sampler.set_sigma_level_upper_limit(
                self._level_sigma_prior.upper_limit)
        if (
                self._slope_sigma_prior.upper_limit > 0
                and np.isfinite(self._slope_sigma_prior.upper_limit)
        ):
            sampler.set_sigma_slope_upper_limit(
                self._slope_sigma_prior.upper_limit)
        self._state_model.set_method(sampler)

    def _set_initial_distribution(self):
        """
        A utility called by the constructor.  See the __init__ method for
        argument documentation.
        """
        initial_state_mean = np.array([
            self._initial_level_prior.mean,
            self._initial_slope_prior.mean,
        ])
        self._state_model.set_initial_state_mean(
            boom.Vector(initial_state_mean))
        initial_sd = np.array([
            self._initial_level_prior.sd,
            self._initial_slope_prior.sd,
        ])
        initial_variance = np.diag(initial_sd * initial_sd)
        self._state_model.set_initial_state_variance(
            boom.SpdMatrix(initial_variance))

    @staticmethod
    def _compute_sdy(sdy, y, which_prior):
        """
        Return the standard deviation of y, computing it if and only if needed.
        """
        if sdy is None:
            if y is None:
                raise Exception(
                    f"One of y, sdy, or {which_prior} must be specified.")
            else:
                sdy = np.nanstd(y, ddof=1)
        return sdy

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()



import numpy as np
import pandas as pd
import BayesBoom.boom as boom
import BayesBoom.R as R


class RegressionSpikeSlabPrior:
    """
    Components of spike and slab priors that are shared regardless of the model
    type.
    """
    def __init__(self,
                 x,
                 y=None,
                 expected_r2=.5,
                 prior_df=.01,
                 expected_model_size=1,
                 prior_information_weight=.01,
                 diagonal_shrinkage=.5,
                 optional_coefficient_estimate=None,
                 max_flips=-1,
                 mean_y=None,
                 sdy=None,
                 prior_inclusion_probabilities=None,
                 sigma_upper_limit=None):
        """
        Computes information that is shared by the different implementation of
        spike and slab priors.  Currently, the only difference between the
        different priors is the prior variance on the regression coefficients.
        When that changes, change this class accordingly, and change all the
        classes that inherit from it.

        Args:
          number_of_variables: The number of columns in the design matrix for
            the regression begin modeled.  The maximum size of the coefficient
            vector.

          expected_r2: The R^2 statistic that the model is expected
            to achieve.  Used along with 'sdy' to derive a prior distribution
            for the residual variance.

          prior_df: The number of observations worth of weight to give to the
            guess at the residual variance.

          expected_model_size: The expected number of nonzero coefficients in
            the model.  Used to set prior_inclusion_probabilities to
            expected_model_size / number_of_variables.  If expected_model_size
            is either negative or larger than number.of.variables then all
            elements of prior_inclusion_probabilities will be set to 1.0 and
            the model will be fit with all available coefficients.

          optional_coefficient_estimate: A vector of length number.of.variables
            to use as the prior mean of the regression coefficients.  This can
            also be None, in which case the prior mean for the intercept will
            be set to mean.y, and the prior mean for all slopes will be 0.

          mean.y: The mean of the response variable.  Used to create a sensible
            default prior mean for the regression coefficients when
            optional_coefficient_estimate is None.

          sdy: Used along with expected_r2 to create a prior guess at the
            residual variance.

          prior_inclusion_probabilities: A vector of length number.of.variables
            giving the prior inclusion probability of each coefficient.  Each
            element must be between 0 and 1, inclusive.  If left as None then a
            default value will be created with all elements set to
            expected_model_size / number_of_variables.

          sigma_upper_limit: The largest acceptable value for the residual
            standard deviation.
        """
        if isinstance(x, pd.DataFrame):
            all_numeric = np.all(x.dtypes.apply(pd.api.types.is_numeric_dtype))
            if all_numeric:
                x = boom.Matrix(x.values)
            else:
                raise Exception("A data frame that included non-numeric data "
                                "was passed to RegressionSpikeSlabPrior.")
        elif isinstance(x, np.ndarray):
            x = boom.Matrix(x)
        if not isinstance(x, boom.Matrix):
            raise Exception(
                "x should either be a 2-dimensional np.array or a boom.Matrix.")

        if mean_y is None:
            if y is None:
                raise Exception("Either 'y' or 'mean_y' must be specified.")
            if isinstance(y, pd.Series):
                y = boom.Vector(y.values)
            elif isinstance(y, np.ndarray):
                y = boom.Vector(y)
            mean_y = boom.mean(y)
        if optional_coefficient_estimate is None:
            optional_coefficient_estimate = np.zeros(x.ncol)
            optional_coefficient_estimate[0] = mean_y
        self._mean = boom.Vector(optional_coefficient_estimate)

        sample_size = x.nrow
        ods = 1. - diagonal_shrinkage
        scale_factor = prior_information_weight * ods / sample_size
        self._unscaled_prior_precision = x.inner() * scale_factor
        diag_view = self._unscaled_prior_precision.diag()
        diag_view /= ods

        if prior_inclusion_probabilities is None:
            potential_nvars = x.ncol
            prob = expected_model_size / potential_nvars
            if prob > 1:
                prob = 1
            if prob < 0:
                prob = 0
            self._prior_inclusion_probabilities = boom.Vector(
                np.full(potential_nvars, prob))
        else:
            self._prior_inclusion_probabilities = boom.Vector(
                prior_inclusion_probabilities)

        if sdy is None:
            sdy = boom.sd(y)
        sample_variance = sdy**2
        expected_residual_variance = (1 - expected_r2) * sample_variance
        self._residual_precision_prior = boom.ChisqModel(
            prior_df,
            np.sqrt(expected_residual_variance))

        if sigma_upper_limit is None:
            self._sigma_upper_limit = sdy * 1.2
        else:
            self._sigma_upper_limit = sigma_upper_limit

        self._max_flips = max_flips

    def __getstate__(self):
        ans = self.__dict__.copy()
        if hasattr(self, "_residual_precision_prior"):
            prior = self._residual_precision_prior
            ans["prior_df"] = 2.0 * prior.alpha()
            ans["prior_ss"] = 2.0 * prior.beta()
        del ans["_residual_precision_prior"]
        return ans

    def __setstate__(self, payload):
        self.__dict__.update(payload)
        self._residual_precision_prior = boom.ChisqModel(
            self.prior_df, np.sqrt(self.prior_ss / self.prior_df))
        del self.prior_df
        del self.prior_ss

    def slab(self, boom_sigsq_prm):
        """Return a boom.MvnGivenScalarSigma model.

        Args:
          boom_sigsq_prm: A BOOM::Ptr<UnivParams> to the residual variance
            parameter for the regression model.

        Returns:
          A BOOM::Ptr<MvnGivenScalarSigma> model that can serve as the slab in
          a spike and slab regression model.

        """
        return boom.MvnGivenScalarSigma(
            self._mean, self._unscaled_prior_precision, boom_sigsq_prm)

    @property
    def spike(self):
        return boom.VariableSelectionPrior(self._prior_inclusion_probabilities)

    @property
    def residual_precision(self):
        """
        A boom.ChisqModel giving the prior distribution for the residual
        precision parameter.
        """
        return self._residual_precision_prior

    @property
    def sigma_upper_limit(self):
        return self._sigma_upper_limit

    @property
    def max_flips(self):
        return self._max_flips


class StudentSpikeSlabPrior(RegressionSpikeSlabPrior):
    """
    A SpikeSlabPrior appropriate for regression models with Student T errors.
    """
    def __init__(
            self,
            x,
            y=None,
            expected_r2=.5,
            prior_df=.01,
            expected_model_size=1,
            prior_information_weight=.01,
            diagonal_shrinkage=.5,
            optional_coefficient_estimate=None,
            max_flips=-1,
            mean_y=None,
            sdy=None,
            prior_inclusion_probabilities=None,
            sigma_upper_limit=np.Inf,
            tail_thickness_prior=R.UniformPrior(0.1, 100)
    ):
        """
        Args:
          tail_thickness_prior: An object with a boom() method, which returns a
            boom.DoubleModel describing the prior on the tail thickness
            parameter.

          All other arguments are as documented in RegressionSpikeSlabPrior.
        """
        super().__init__(
            x=x,
            y=y,
            expected_r2=expected_r2,
            prior_df=prior_df,
            expected_model_size=expected_model_size,
            prior_information_weight=prior_information_weight,
            diagonal_shrinkage=diagonal_shrinkage,
            optional_coefficient_estimate=optional_coefficient_estimate,
            max_flips=max_flips,
            mean_y=mean_y,
            sdy=sdy,
            prior_inclusion_probabilities=prior_inclusion_probabilities,
            sigma_upper_limit=sigma_upper_limit)
        self._nu_prior = tail_thickness_prior

    @property
    def tail_thickness(self):
        """
        A boom.DoubleModel giving the prior distribution on the tail thickness
        parameter.
        """
        return self._nu_prior.boom()

    def create_sampler(self, model, assign=False):
        pass


def logit(p):
    ans = np.log(p / (1 - p))
    return float(ans)


class LogitZellnerPrior:
    def __init__(self,
                 predictors,
                 successes=None,
                 trials=None,
                 prior_success_probability=0.5,
                 expected_model_size=1.0,
                 prior_information_weight=1.0,
                 diagonal_shrinkage=.5,
                 optional_coefficient_estimate=None,
                 max_flips=-1,
                 prior_inclusion_probabilities=None):
        """
        Args:
          predictors: A matrix of predictor variables, denoted X below.
          successes:  A vector of success counts.
          trials: A vector of trial counts.  If successes and trials are given,
            they must satisfy 0 <= successes <= trials.  If successes is given
            and trials is None, then successes must only contain 0's and 1's.

          prior_success_probability: If 'successes' is not given,
            then prior_success_probability is used to scale the prior mean of
            the intercept term.  Otherwise
        """

        self._max_flips = max_flips
        sample_size = predictors.shape[0]
        xdim = predictors.shape[1]

        xtx = predictors.T @ predictors * prior_information_weight / sample_size
        xtx_diagonal = np.diagonal(xtx).copy()
        xtx *= 1 - diagonal_shrinkage
        np.fill_diagonal(xtx, xtx_diagonal)

        self._precision = xtx
        self._mean = np.zeros(xdim)
        if successes is None:
            self._mean[0] = logit(prior_success_probability)
        else:
            if trials is None:
                trials = np.ones(sample_size)
            p_hat = np.nanmean(successes / trials)
            self._mean[0] = logit(p_hat)
        if not np.isfinite(self._mean[0]):
            self._mean[0] = 0.0

        if prior_inclusion_probabilities is None:
            prior_inclusion_probabilities = np.full(
                xdim, expected_model_size / xdim)
        self._prior_inclusion_probabilities = prior_inclusion_probabilities

    @classmethod
    def from_parameters(self, mean, precision, prior_inclusion_probabilities,
                        max_flips=-1):
        xdim = len(mean)
        predictors = np.random.randn(xdim, xdim)
        y = np.full(xdim, 0.5)
        trials = np.ones(xdim)
        ans = LogitZellnerPrior(predictors, y, trials, max_flips=max_flips)
        ans._prior_inclusion_probabilities = prior_inclusion_probabilities
        ans._mean = mean
        ans._precision = precision
        return ans

    @property
    def slab(self):
        return boom.MvnModel(
            boom.Vector(self._mean),
            boom.SpdMatrix(self._precision),
            True)

    @property
    def spike(self):
        return boom.VariableSelectionPrior(
            self._prior_inclusion_probabilities)

    @property
    def max_flips(self):
        return self._max_flips

    def create_sampler(self, model, assign=False):
        if not isinstance(model, boom.BinomialLogitModel):
            raise Exception("Expected 'model' to be a boom.BinomialLogitModel.")
        sampler = boom.BinomialLogitSpikeSlabSampler(
            model=model,
            slab=self.slab,
            spike=self.spike,
            clt_threshold=5,
            seeding_rng=boom.GlobalRng.rng)
        if self._max_flips > 0 and np.isfinite(self._max_flips):
            sampler.limit_model_selection(self._max_flips)
        if assign:
            model.set_method(sampler)
        return sampler


class PoissonZellnerPrior:
    def __init__(self,
                 predictors,
                 counts=None,
                 exposure=None,
                 prior_event_rate=1.0,
                 expected_model_size=1.0,
                 prior_information_weight=1.0,
                 diagonal_shrinkage=.5,
                 optional_coefficient_estimate=None,
                 max_flips=-1,
                 prior_inclusion_probabilities=None):
        self._max_flips = max_flips
        sample_size = predictors.shape[0]
        xdim = predictors.shape[1]

        xtx = predictors.T @ predictors * prior_information_weight / sample_size
        xtx_diagonal = np.diagonal(xtx).copy()
        xtx *= 1 - diagonal_shrinkage
        np.fill_diagonal(xtx, xtx_diagonal)

        self._precision = xtx
        self._mean = np.zeros(xdim)
        if counts is None:
            self._mean[0] = np.log(prior_event_rate)
        else:
            if exposure is None:
                exposure = np.ones(sample_size)
            p_hat = np.nanmean(counts / exposure)
            self._mean[0] = logit(p_hat)
        if not np.isfinite(self._mean[0]):
            self._mean[0] = 0.0

        if prior_inclusion_probabilities is None:
            prior_inclusion_probabilities = np.full(
                xdim, expected_model_size / xdim)
        self._prior_inclusion_probabilities = prior_inclusion_probabilities

    @classmethod
    def from_parameters(self, mean, precision, prior_inclusion_probabilities,
                        max_flips=-1):
        xdim = len(mean)
        predictors = np.random.randn(xdim, xdim)
        y = np.full(xdim, 0.5)
        trials = np.ones(xdim)
        ans = LogitZellnerPrior(predictors, y, trials, max_flips=max_flips)
        ans._prior_inclusion_probabilities = prior_inclusion_probabilities
        ans._mean = mean
        ans._precision = precision
        return ans

    @property
    def slab(self):
        return boom.MvnModel(
            boom.Vector(self._mean),
            boom.SpdMatrix(self._precision),
            True)

    @property
    def spike(self):
        return boom.VariableSelectionPrior(
            self._prior_inclusion_probabilities)

    @property
    def max_flips(self):
        return self._max_flips

    def create_sampler(self, model, assign=False):
        if not isinstance(model, boom.PoissonRegressionModel):
            raise Exception(
                "Expected 'model' to be a boom.PoissonRegressionModel.")
        sampler = boom.PoissonRegressionSpikeSlabSampler(
            model=model,
            slab=self.slab,
            spike=self.spike,
            clt_threshold=5,
            seeding_rng=boom.GlobalRng.rng)
        if self._max_flips > 0 and np.isfinite(self._max_flips):
            sampler.limit_model_selection(self._max_flips)
        if assign:
            model.set_method(sampler)
        return sampler



import unittest
import numpy as np
import pandas as pd
import pickle

import pdb

import matplotlib.pyplot as plt

import BayesBoom.R as R

from BayesBoom.spikeslab import dot

from BayesBoom.bsts import (
    Bsts,
    AirPassengers,
    compare_bsts_models,
    LocalLevelStateModel,
    LocalLinearTrendStateModel,
    SeasonalStateModel,
)


class TestGaussianTimeSeries(unittest.TestCase):
    def setUp(self):
        n = 100
        random_walk = np.cumsum(np.random.randn(n) * .1)
        noise = np.random.randn(n)
        self.data = random_walk + noise

    def tearDown(self):
        R.delete_if_present("bsts_llt.pkl")

    def test_local_level(self):
        model = Bsts()
        model.add_state(LocalLevelStateModel(self.data))
        model.train(data=self.data, niter=10)

        with open("bsts_llt.pkl", "wb") as pkl:
            pickle.dump(model, pkl)

        with open("bsts_llt.pkl", "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertIsInstance(m2, Bsts)
        np.testing.assert_array_equal(m2._final_state, model._final_state)

        errors = model.one_step_prediction_errors()
        self.assertIsInstance(errors, np.ndarray)
        self.assertEqual(errors.shape, (10, 100))

        errors = model.one_step_prediction_errors(burn=7)
        self.assertEqual(errors.shape, (3, 100))

        foo = model._model.simulate_holdout_prediction_errors(
            50, 90, False).to_numpy()
        self.assertEqual(foo.shape, (50, 100))
        self.assertEqual(100, model.time_dimension)

        cutpoints = [60, 80, 100]
        errors = model.one_step_prediction_errors(cutpoints=cutpoints)
        self.assertIsInstance(errors, dict)
        self.assertEqual(len(errors), len(cutpoints))
        self.assertEqual(errors[60].shape,  (10, 100))
        self.assertEqual(errors[80].shape,  (10, 100))
        self.assertEqual(errors[100].shape,  (10, 100))
        self.assertEqual(model.time_dimension, 100)


    def test_basic_structural_model(self):
        model = Bsts()

        y = np.log(AirPassengers)
        model.add_state(LocalLinearTrendStateModel(y))
        model.add_state(SeasonalStateModel(y, nseasons=12))
        model.train(data=y, niter=1000)

        fig = model.plot("comp")
        if _show_figs:
            fig.show()

        predictions = model.predict(12)
        target_quantiles = [0.025] + list(np.linspace(.05, 0.95, 19)) + [0.975]
        prediction_quantiles = np.quantile(
            predictions.distribution, target_quantiles, axis=0)
        self.assertIsInstance(prediction_quantiles, np.ndarray)

        with open("bsm.pkl", "wb") as pkl:
            pickle.dump(model, pkl)

        with open("bsm.pkl", "rb") as pkl:
            m2 = pickle.load(pkl)

        pred2 = m2.predict(12)

        np.testing.assert_array_almost_equal(predictions.distribution,
                                             pred2.distribution)

        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        predictions.plot(ax=ax[0], original_series=24)
        pred_plot_fig, pred_plot_ax = pred2.plot(ax=ax[1], original_series=24)
        if _show_figs:
            fig.show()
        self.assertIsInstance(pred_plot_ax, plt.Axes)


class TestGaussianRegression(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        n = 100
        p = 10
        pgood = 3
        X = np.random.randn(n, p)
        beta = np.zeros(p)
        beta[:pgood] = np.random.randn(pgood)
        yhat = X @ beta

        random_walk = np.cumsum(np.random.randn(n) * .1)
        noise = np.random.randn(n) * .05
        y = yhat + random_walk + noise
        self.data = pd.DataFrame(X, columns=R.paste0("X", np.arange(1, p + 1)))
        self.data["y"] = y

    def test_state_space_regression(self):
        pass


class TestStudentTimeSeries(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        n = 100
        df = 3
        random_walk = np.cumsum(np.random.randn(n) * .1)
        noise = np.random.standard_t(df=df, size=n)
        self.data = random_walk + noise

    def tearDown(self):
        R.delete_if_present("bsts_student_llt.pkl")
        R.delete_if_present("bsm.pkl")

    def test_local_level(self):
        model = Bsts(family="student")
        model.add_state(LocalLevelStateModel(self.data))
        model.train(data=self.data, niter=1000)

        fname = "bsts_student_llt.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(model, pkl)

        with open(fname, "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertEqual(model.time_dimension, m2.time_dimension)
        self.assertIsInstance(m2, Bsts)

        np.testing.assert_array_almost_equal(
            model._observation_model_manager._residual_sd,
            m2._observation_model_manager._residual_sd)
        np.testing.assert_array_almost_equal(
            model._observation_model_manager._residual_df,
            m2._observation_model_manager._residual_df)

        np.testing.assert_array_almost_equal(
            model._state_models[0].state_contribution,
            m2._state_models[0].state_contribution)
        np.testing.assert_array_almost_equal(
            model._state_models[0].sigma_draws,
            m2._state_models[0].sigma_draws)

        np.testing.assert_array_almost_equal(
            model._final_state, m2._final_state)

        seed = 8675309
        pred1 = model.predict(4, seed=seed)
        pred2 = m2.predict(4, seed=seed)

        fig, ax = plt.subplots(1, 2)
        pred1.plot(ax=ax[0])
        pred2.plot(ax=ax[1])
        if _show_figs:
            fig.show()

        np.testing.assert_array_almost_equal(
            pred1.distribution,
            pred2.distribution)

    def test_basic_structural_model(self):
        model = Bsts(family="student")

        y = np.log(AirPassengers)
        model.add_state(LocalLinearTrendStateModel(y))
        model.add_state(SeasonalStateModel(y, nseasons=12))
        model.train(data=y, niter=1000)

        fig = model.plot("comp")
        if _show_figs:
            fig.show()

        horizon = 24
        predictions = model.predict(horizon)
        target_quantiles = [0.025] + list(np.linspace(.05, 0.95, 19)) + [0.975]
        prediction_quantiles = np.quantile(
            predictions.distribution, target_quantiles, axis=0)
        self.assertIsInstance(prediction_quantiles, np.ndarray)

        with open("bsm.pkl", "wb") as pkl:
            pickle.dump(model, pkl)

        with open("bsm.pkl", "rb") as pkl:
            m2 = pickle.load(pkl)

        pred2 = m2.predict(horizon)

        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        predictions.plot(ax=ax[0], original_series=60)
        pred_plot_fig, pred_plot_ax = pred2.plot(ax=ax[1], original_series=60)
        if _show_figs:
            fig.show()
        self.assertIsInstance(pred_plot_ax, plt.Axes)

        np.testing.assert_array_almost_equal(predictions.distribution,
                                             pred2.distribution)

# class TestStateSpaceRegression(unittest.TestCase):

#     def test_bsts_regression(self):
#         sample_size = 100
#         random_walk = np.cumsum(np.random.randn(sample_size) * .05)
#         noise = np.random.randn(sample_size) * .2
#         predictors = pd.DataFrame(np.random.randn(sample_size, 2),
#                                   columns=["x1", "x2"])
#         y = random_walk + 2.3 * predictors["x1"] -1.8 * predictors["x2"] + noise
#         predictors["y"] = y

#         model = Bsts()
#         model.add_state(LocalLevelStateModel(y))
#         model.train(data=predictors, formula="y ~ x1 + x2", niter=100)

#         with open("bsts_reg.pkl", "wb") as pkl:
#             pickle.dump(model, pkl)

#         with open("bsts_reg.pkl", "rb") as pkl:
#             m2 = pickle.load(pkl)


class TestPlots(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        np.random.seed(8675309)
        np.random.seed(0xDEADBEEF)
        fake_data = np.random.randn(100, 10)
        fake_coefficients = np.array([5, -4, 3] + [0] * 7)
        random_walk = np.cumsum(np.random.randn(100))
        noise = np.random.randn(100) * 1.5
        y = random_walk + fake_data @ fake_coefficients + noise
        data = pd.DataFrame(
            fake_data,
            columns=["V" + str(x+1) for x in range(10)],
        )
        data["y"] = y
        model = Bsts()
        model.add_state(LocalLevelStateModel(y))
        model.train(formula=f"y ~ {dot(data, omit='y')}", niter=100, data=data)
        self._regression_model = model

        data = np.log(AirPassengers)
        model = Bsts()
        model.add_state(LocalLinearTrendStateModel(data))
        model.add_state(SeasonalStateModel(data, nseasons=12))
        model.train(data, niter=100)
        self._model = model

        model = Bsts()
        model.add_state(LocalLinearTrendStateModel(data))
        model.train(data, niter=100)
        self._model2 = model

    def test_plot_state(self):
        _, ax = plt.subplots()
        foo, bar = self._model.plot(ax=ax)
        self.assertIsInstance(bar, plt.Axes)

        _, ax = plt.subplots()
        foo, bar = self._regression_model.plot(ax=ax)
        self.assertIsInstance(bar, plt.Axes)

    def test_plot_components(self):
        fig = plt.figure()
        foo = self._model.plot("comp", fig=fig)
        self.assertIsInstance(foo, plt.Figure)

    def test_plot_seasonal(self):
        fig = plt.figure()
        foo = self._model.plot("seas", fig=fig)
        self.assertIsInstance(foo, plt.Figure)

    def test_plot_residuals(self):
        fig, ax = plt.subplots()
        foo = self._model.plot_residuals(ax=ax)
        self.assertIsInstance(foo, plt.Axes)

        fig, ax = plt.subplots()
        foo = self._regression_model.plot_residuals(ax=ax)
        self.assertIsInstance(foo, plt.Axes)

        if _show_figs:
            fig.show()

    def test_plot_forecast_distribution(self):
        fig = self._model.plot("forecast_distribution", cex_actuals=1)
        self.assertIsInstance(fig, plt.Figure)
        if _show_figs:
            fig.show()

        fig2 = self._model.plot("forecast_distribution",
                                cutpoints=(50, 80, 100))
        if _show_figs:
            fig2.show()

    def test_plot_size(self):
        fig, ax = plt.subplots()
        foo = self._regression_model.plot_size(ax=ax)
        if _show_figs:
            fig.show()
        self.assertIsInstance(foo, plt.Axes)

    def test_plot_predictors(self):
        fig, ax = plt.subplots()
        foo = self._regression_model.plot(
            "predictors", ax=ax, short_names=False)
        if _show_figs:
            fig.show()
        self.assertIsInstance(foo, plt.Axes)

    def test_compare_bsts_models(self):
        fig = plt.figure()
        models = {
            "bsm": self._model,
            "trend only": self._model2
        }
        ans = compare_bsts_models(models, fig=fig)
        self.assertIsInstance(ans, plt.Figure)


_debug_mode = True
_show_figs = _debug_mode

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
#    import warnings
#    warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestPlots()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_plot_forecast_distribution()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import numpy as np
import scipy.stats as stats


def simulate_random_walk(nsteps, sd=0.1):
    return np.cumsum(np.random.randn(nsteps) * sd)


def simulate_student_random_walk(nsteps, sd=0.1, nu=3):
    errors = stats.t.rvs(df=nu, scale=sd, size=nsteps)
    return np.cumsum(errors)


def simulate_local_linear_trend(nsteps, level_sd=0.1, slope_sd=0.05, mu0=0):
    slope = simulate_random_walk(nsteps, sd=slope_sd)
    error = np.random.randn(nsteps) * level_sd

    slopes = np.array([0.0] + slope[:-1].tolist())
    errors = np.array([0.0] + error[:-1].tolist())
    trend = mu0 + np.cumsum(slopes) + np.cumsum(errors)
    return trend


def simulate_student_local_linear_trend(nsteps, level_sd=0.1, level_df=3,
                                        slope_sd=0.05, slope_df=5, mu0=0):
    slope = simulate_student_random_walk(nsteps, sd=slope_sd, nu=slope_df)
    error = stats.t.rvs(size=nsteps, scale=level_sd, df=level_df)

    slopes = np.array([0.0] + slope[:-1].tolist())
    errors = np.array([0.0] + error[:-1].tolist())
    trend = mu0 + np.cumsum(slopes) + np.cumsum(errors)
    return trend



import numpy as np
import BayesBoom.boom as boom
import BayesBoom.R as R
from .state_models import StateModel


class TrigStateModel(StateModel):
    """
    Describe seasonal effects with harmonic functions rather than dummy
    variables.
    """

    def __init__(self, y, period, frequencies, sigma_prior=None,
                 initial_state_prior=None, sdy=None):
        """
        Args:
          y:  The time series to be modeled.
          period:  The number of time steps in a full cycle.
          frequencies: A vector of positive real numbers, giving the number of
            times a cycle repeats in a period.  One sine and one cosine term
            will be added to the state for each frequency.
          sigma_prior: The prior distribution for the standard deviations of
            the changes in the sinusoid coefficients at each new time point.
            This can be None (in which case a default prior will be used), or a
            single object of class R.SdPrior (which will be repeated for each
            sinusoid independently).
          initial_state_prior: The prior distribution for the values of the
            sinusoid coefficients at time 0.  This can either be None (in which
            case a default prior will be used), or an object of class
            R.MvnPrior.  If the prior is specified directly its dimension must
            be twice the number of frequencies.
          sdy: The standard deviation of the time series to be modeled.  This
            argument is ignored if y is provided.
        """
        self._period = period
        self._frequencies = frequencies
        if sdy is None:
            sdy = np.nanstd(y)

        self._validate_prior(sigma_prior, sdy)
        self._validate_initial_distribution(initial_state_prior, sdy)
        self._state_contribution = None
        self._build_state_model()

    @property
    def label(self):
        ans = "Trig["
        ans += f"{self._period:.2f}" + " | "
        for freq in self._frequencies:
            ans += f"{freq:.1f}, "
        ans = ans[:-2] + "]"
        return ans

    @property
    def state_dimension(self):
        return 2 * len(self._frequencies)

    @property
    def state_error_dimension(self):
        return self.state_dimension

    @property
    def state_contribution(self):
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self._sigma = np.empty(niter)
        self._state_contribution = np.empty((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self._sigma[iteration] = self._state_model.error_distribution.sd
        begin = self._state_index
        end = begin + self.state_dimension
        state = state_matrix[begin:end, :]

        self._state_contribution[iteration, :] = (
            self._state_model.compute_state_contribution(state)
        ).to_numpy()

    def restore_state(self, iteration):
        self._state_model.error_distribution.set_sigma(
            self._sigma[iteration])

    def _validate_prior(self, sigma_prior, sdy):
        if sigma_prior is None:
            sigma_prior = R.SdPrior(sdy * .01, upper_limit=sdy)
        if not isinstance(sigma_prior, R.SdPrior):
            raise Exception("Wrong type for sigma_prior.")
        self._sigma_prior = sigma_prior

    def _validate_initial_distribution(self, initial_state_prior, sdy):
        if initial_state_prior is None:
            dim = self.state_dimension
            initial_state_prior = R.MvnPrior(
                np.zeros(dim),
                np.diag(np.ones(dim) * sdy ** 2))
        if not isinstance(initial_state_prior, R.MvnPrior):
            raise Exception("Wrong type for initial_state_prior.")
        if len(initial_state_prior.mean) != self.state_dimension:
            raise Exception(
                f"Initial_state_prior dimension was {initial_state_prior.dim}."
                f"  State dimension is {self.state_dimension}."
            )
        self._initial_state_prior = initial_state_prior

    def _build_state_model(self):
        self._state_model = boom.TrigStateModel(
            self._period, R.to_boom_vector(self._frequencies))
        self._set_posterior_sampler()
        self._set_initial_distribution()

    def _set_posterior_sampler(self):
        sampler = boom.ZeroMeanGaussianConjSampler(
            self._state_model.error_distribution,
            self._sigma_prior.boom(),
            boom.GlobalRng.rng)
        if (
                self._sigma_prior.upper_limit > 0
                and np.isfinite(self._sigma_prior.upper_limit)
        ):
            sampler.set_sigma_upper_limit(self._sigma_prior.upper_limit)
        self._state_model.set_method(sampler)

    def _set_initial_distribution(self):
        self._state_model.set_initial_state_mean(
            R.to_boom_vector(self._initial_state_prior.mean))
        self._state_model.set_initial_state_variance(
            R.to_boom_spd(self._initial_state_prior.variance))

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()



import unittest
import numpy as np
import os
import pickle

from BayesBoom.bsts import (
    LocalLevelStateModel,
    LocalLinearTrendStateModel,
    SeasonalStateModel,
    SemilocalLinearTrendStateModel
)


class TestLocalLevel(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        self.data = np.random.randn(100)
        self.model = LocalLevelStateModel(self.data)
        self.model.set_state_index(0)

    def tearDown(self):
        if os.path.exists("ll.pkl") and os.path.isfile("ll.pkl"):
            os.remove("ll.pkl")

    def test_state_dimension(self):
        self.assertEqual(1, self.model.state_dimension)

    def test_storage(self):
        self.model.allocate_space(10, 20)
        self.assertEqual(self.model.sigma_draws.shape, (10,))
        self.assertEqual(self.model.state_contribution.shape, (10, 20))

        state_matrix = np.zeros((4, 20))
        self.model.set_state_index(0)
        self.model._state_model.set_sigsq(9.0)

        iteration = 7
        self.model.record_state(iteration, state_matrix)
        self.assertAlmostEqual(self.model.sigma_draws[iteration], 3.0)

        self.model._state_model.set_sigsq(1.0)
        self.model.restore_state(iteration)
        self.assertAlmostEqual(self.model._state_model.sigsq, 9.0)

        with open("ll.pkl", "wb") as pkl:
            pickle.dump(self.model, pkl)

        with open("ll.pkl", "rb") as pkl:
            model = pickle.load(pkl)

        self.assertIsInstance(model, LocalLevelStateModel)

        np.testing.assert_array_equal(
            model.sigma_draws,
            self.model.sigma_draws)


class TestLocalLinearTrend(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        self.data = np.random.randn(100)
        self.model = LocalLinearTrendStateModel(self.data)
        self.model.set_state_index(1)

    def tearDown(self):
        if os.path.exists("llt.pkl") and os.path.isfile("llt.pkl"):
            os.remove("llt.pkl")

    def test_state_dimension(self):
        self.assertEqual(2, self.model.state_dimension)

    def test_storage(self):
        time_dimension = 20
        niter = 10
        self.model.allocate_space(niter, time_dimension)
        self.assertEqual(self.model.sigma_level.shape, (niter,))
        self.assertEqual(self.model.sigma_slope.shape, (niter,))
        self.assertEqual(self.model.state_contribution.shape,
                         (niter, time_dimension))

        iteration = 4
        self.model.sigma_level[iteration] = 2.6
        self.model.sigma_slope[iteration] = 1.9
        self.model.restore_state(iteration)
        self.assertAlmostEqual(self.model._state_model.sigma_level, 2.6)
        self.assertAlmostEqual(self.model._state_model.sigma_slope, 1.9)

        state_matrix = np.random.randn(3, time_dimension)
        iteration = 7
        self.model.record_state(iteration, state_matrix)
        self.assertAlmostEqual(self.model.sigma_level[iteration], 2.6)
        self.assertAlmostEqual(self.model.sigma_slope[iteration], 1.9)
        np.testing.assert_array_equal(
            self.model._state_contribution[iteration, :],
            state_matrix[1, :])

        with open("llt.pkl", "wb") as pkl:
            pickle.dump(self.model, pkl)

        with open("llt.pkl", "rb") as pkl:
            m2 = pickle.load(pkl)

        np.testing.assert_array_equal(self.model.sigma_level, m2.sigma_level)
        np.testing.assert_array_equal(self.model.sigma_slope, m2.sigma_slope)


class TestSeasonalStateModel(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        self.data = np.random.randn(100)
        self.model = SeasonalStateModel(self.data, nseasons=4)
        self.model.set_state_index(1)

    def tearDown(self):
        if os.path.exists("seas.pkl") and os.path.isfile("seas.pkl"):
            os.remove("seas.pkl")

    def test_state_dimension(self):
        self.assertEqual(3, self.model.state_dimension)

    def test_storage(self):
        time_dimension = 20
        niter = 10
        self.model.allocate_space(niter, time_dimension)
        self.assertEqual(self.model.sigma_draws.shape, (niter, ))
        self.assertEqual(self.model._state_contribution.shape,
                         (niter, time_dimension))

        self.model.set_state_index(2)
        state_matrix = np.random.randn(3, time_dimension)
        iteration = 4
        self.model.sigma_draws[4] = 1.9
        self.model.restore_state(iteration)
        self.assertAlmostEqual(1.9, self.model._state_model.sigma)

        iteration = 8
        self.model.record_state(iteration, state_matrix)
        self.assertAlmostEqual(self.model.sigma_draws[iteration], 1.9)
        np.testing.assert_array_equal(
            self.model._state_contribution[iteration, :],
            state_matrix[self.model._state_index, :])

        with open("seas.pkl", "wb") as pkl:
            pickle.dump(self.model, pkl)

        with open("seas.pkl", "rb") as pkl:
            _ = pickle.load(pkl)


class TestSemilocalLinearTrendStateModel(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)
        self.data = np.random.randn(100)
        self.model = SemilocalLinearTrendStateModel(self.data)
        self.model.set_state_index(0)

    def tearDown(self):
        if os.path.exists("sllt.pkl") and os.path.isfile("sllt.pkl"):
            os.remove("sllt.pkl")

    def test_state_dimension(self):
        self.assertEqual(self.model.state_dimension, 3)

    def test_storage(self):
        time_dimension = 20
        niter = 10
        self.model.allocate_space(niter, time_dimension)
        self.assertEqual(self.model.level_sigma.shape, (niter,))
        self.assertEqual(self.model.slope_sigma.shape, (niter,))
        self.assertEqual(self.model.slope_ar1.shape, (niter,))
        self.assertEqual(self.model.slope_mean.shape, (niter,))
        self.assertEqual(self.model.state_contribution.shape,
                         (niter, time_dimension))

        iteration = 4
        self.model.level_sigma[iteration] = 2.6
        self.model.slope_sigma[iteration] = 1.9
        self.model.slope_mean[iteration] = -7.2
        self.model.slope_ar1[iteration] = 0.8
        self.model.restore_state(iteration)
        self.assertAlmostEqual(self.model._state_model.level_sd, 2.6)
        self.assertAlmostEqual(self.model._state_model.slope_sd, 1.9)
        self.assertAlmostEqual(self.model._state_model.slope_mean, -7.2)
        self.assertAlmostEqual(self.model._state_model.slope_ar_coefficient,
                               0.8)

        state_matrix = np.random.randn(3, time_dimension)
        iteration = 7
        self.model.record_state(iteration, state_matrix)
        self.assertAlmostEqual(self.model.level_sigma[iteration], 2.6)
        self.assertAlmostEqual(self.model.slope_sigma[iteration], 1.9)
        self.assertAlmostEqual(self.model.slope_mean[iteration], -7.2)
        self.assertAlmostEqual(self.model.slope_ar1[iteration], 0.8)
        np.testing.assert_array_equal(
            self.model._state_contribution[iteration, :],
            state_matrix[self.model._state_index, :])

        with open("sllt.pkl", "wb") as pkl:
            pickle.dump(self.model, pkl)

        with open("sllt.pkl", "rb") as pkl:
            m2 = pickle.load(pkl)

        np.testing.assert_array_equal(self.model.level_sigma, m2.level_sigma)
        np.testing.assert_array_equal(self.model.slope_sigma, m2.slope_sigma)
        np.testing.assert_array_equal(self.model.slope_ar1, m2.slope_ar1)
        np.testing.assert_array_equal(self.model.slope_mean, m2.slope_mean)


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")
    rig = TestSemilocalLinearTrendStateModel()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_storage()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import pandas as pd
import numpy as np
from inspect import isfunction, getsource
import time
from numbers import Number


class omit:
    """
    For use in indexing. my_data[omit(bad_integers), :] omits the bad rows.
    """


class data_frame(pd.DataFrame):
    """A data frame that indexes like R.  self[] takes two entries, rows and
    columns.

    TODO: flesh out the indexer, and check that all the pandas methods still
    work.

    """

    def __init__(self, *args, **kwargs):
        """The constructor is identical to pandas constructor.
        """
        super().__init__(*args, **kwargs)

    def __getattr__(self, name):
        return super().__getitem__(name)

    def __getitem__(self, arg):
        if not isinstance(arg, tuple) and not len(arg) == 2:
            raise Exception("Rows and columns must be supplied.")

        rows = arg[0]
        cols = arg[1]
        print("rows =  ", rows)
        print("cols =  ", cols)
        # TODO actually do the indexing.
        # In most cases indexing by columns first is the right thing to do.

        # If slice and not None, None, None

    @property
    def nrow(self):
        return self.shape[0]

    @property
    def ncol(self):
        return self.shape[1]


def pretty(list_of_strings, width=80, hide_underscore=True):
    '''Print a list of strings in formatted columns.

    Args:
      list_of_strings: The list of strings to print.
      width:  The width of the screen.
      hide_underscore: If True then elements in list_of_strings that begin or
        end with underscores will not be printed.

    Returns:
      None

    Effects:
      The strings are printed to the screen.
    '''
    if hide_underscore:
        private = [(x[0] == '_') | (x[-1] == '_') for x in list_of_strings]
        to_print = [x for x, hide in zip(
            list_of_strings, private) if hide is False]
    else:
        to_print = list_of_strings

    if len(to_print) == 0:
        return
    max_len = np.max([len(x) for x in to_print])
    entry_width = max_len + 2

    line = ''
    for entry in to_print:
        # This would be a good place to use string formatting, but I had
        # trouble getting it to work.
        buffer = ' ' * (entry_width - len(entry))
        padded = entry + buffer
        if len(line) + entry_width <= width:
            line += padded
        else:
            print(line)
            line = padded
    print(line)


def print_timestamp(iteration_number, ping):
    if ping <= 0:
        return
    if iteration_number % ping == 0:
        timestamp = time.asctime()
        sep = ' =-=-=-=-=-=-=-=-=-=-= '
        print(sep + timestamp + f" Iteration {iteration_number} " + sep)


def ls(*args, hide_underscore=True):
    """
    List the contents of one or more objects.  If passed a function then print
    the body of the function.
    """

    if len([*args]) == 0:
        print("If you're trying to list the global namespace type "
              "'pretty(dir())'.")
        print("(The interactive namespace is not available to modules.)")
#        pretty(sorted(globals()), hide_underscore=hide_underscore)
    elif len([*args]) == 1 and isfunction(args[0]):
        print(getsource(args[0]))
    else:
        for arg in args:
            pretty(dir(arg), hide_underscore=hide_underscore)
            print("\n")


# Need the formula language so we can ask for conditional distributions.  This
# function is better than pd.value_counts because it handles numpy and list
# data too.
def table(*args):
    """
    Compute a frequency table of one or more categorial variables.
    """
    if len(args) == 1:
        if isinstance(args[0], pd.DataFrame):
            return args[0].crosstab(margins=True)
        else:
            x = pd.Series(args[0])
            return x.value_counts()
    else:
        x = pd.DataFrame(*args)
        return x.crosstab(margins=True)


def data_range(x):
    """
    Return the smallest and largest entries in x.  The name distinguishes this
    function from the python built-in 'range'.
    """
    return np.quantile(x, q=[0, 1])


def corr(*args):
    """
    Compute the correlation among one or more objects.  If a single matrix or
    data frame is passed, or if a collection of vectors is passed, then return
    the correlation matrix.  If a pair of vectors is passed, return the number.

    This function corrects the stupid default in numpy which assumes variables
    are rows rather than columns.
    """
    if len(args) == 1:
        x = args[0]
    else:
        x = np.stack(args, axis=1)
        if x.shape[1] == 2:
            return np.corrcoef(x, rowvar=False)[0, 1]
    return np.corrcoef(x, rowvar=False)


def first_true(boolean_array):
    """
    Returns the index of the first True element in the array-like boolean_array.
    Returns None if no True values are found.
    """
    return next((i for i, v in enumerate(boolean_array) if v), None)


def recycle(x, output_len):
    x = list(x)
    nchoices = len(x)
    if nchoices >= output_len:
        return x[:output_len]
    ans = x
    len_ans = nchoices
    while len_ans + nchoices < output_len:
        ans = ans + x
        len_ans += nchoices
    num_remaining = output_len - len_ans
    return ans + x[:num_remaining]


def unique_match(value, legal_value_list):
    """
    If 'value' uniquely matches only one value in legal_value_list, then return
    the corresponding value in legal_value_list.  If not then return None.
    """

    matches = np.array([x.startswith(value) for x in legal_value_list])
    if matches.sum() != 1:
        return None
    else:
        return legal_value_list[first_true(matches)]


def _reduce_concat(x, sep=""):
    import functools
    return functools.reduce(lambda x, y: str(x) + sep + str(y), x)


def _deduce_type(*lists):
    """
    Return an integer indicating the "most general" type in the variable length
    argument list *lists.  Return type 2 is pd.Series, return type 1 is
    np.array, and return type 0 is list.

    Args:
      *lists: a sequence of vector-like objects.
    """
    type_code = 0
    for obj in lists:
        if isinstance(obj, pd.Series):
            type_code = 2
        elif isinstance(obj, np.ndarray):
            type_code = max(type_code, 1)
    return type_code


def paste(*lists, sep=" ", collapse=None):
    """
    Paste one or more vector-like objects to gether into a vector of strings.

    Args:
      *lists:  A sequence of vector-like objects.

    Returns:
      A vector-like object containing the concatenated string representations
      of the arguments.  If any of the arguments are a pd.Series then the
      return type is also pd.Series.  Otherwise if any arguments are numpy
      arrays the return is a numpy array.  Otherwise the return type is a list.
    """
    list_of_inputs = [*lists]
    max_length = np.max([len(x) for x in list_of_inputs])
    for i in range(len(list_of_inputs)):
        if isinstance(list_of_inputs[i], str):
            string_value = list_of_inputs[i]
            list_of_inputs[i] = [string_value] * max_length
        elif isinstance(list_of_inputs[i], Number):
            value = list_of_inputs[i]
            list_of_inputs[i] = [value] * max_length

    parallel_data = pd.DataFrame(list_of_inputs).T
    result = parallel_data.apply(_reduce_concat, sep=sep, axis=1).astype(
        str).values.tolist()
    if collapse is not None:
        return _reduce_concat(result, sep=collapse)
    else:
        return result


def paste0(*lists, sep="", collapse=None):
    return paste(*lists, sep=sep, collapse=collapse)


def remove_common_prefix(strings):
    if len(strings) == 0:
        return strings
    current = [x for x in strings]
    while True:
        try:
            initial_element = {x[0] for x in current}
        except IndexError:
            return current
        if len(initial_element) > 1:
            return current
        current = [x[1:] for x in current]


def remove_common_suffix(strings):
    """
    If all strings in a collection end in the same character, that character is
    a 'common suffix."  This function reomves common suffixes from collections
    of strings.

    Args:
      strings:  A list, array, or other collection of strings.

    Returns:
      shortened_strings:  A list of string with common suffixes removed.

    Examples:
      remove_common_suffix([])
      []

      remove_common_suffix(["foo", "bar", "baz"])
      ["foo", "bar", "baz"]

      remove_common_suffix(["fooz", "barz", "baz"])
      ["foo", "bar", "ba"]

      remove_common_suffix(["zz", "z"])
      ["z", ""]

    """
    if len(strings) == 0:
        return strings
    current = [x for x in strings]
    while True:
        try:
            final_element = {x[-1] for x in current}
        except IndexError:
            return current
        if len(final_element) > 1:
            return current
        current = [x[:-1] for x in current]



import unittest
import BayesBoom.R as R
import numpy as np
import pandas as pd


class TestRfuns(unittest.TestCase):

    def setUp(self):
        np.random.seed(8675309)

    def test_paste(self):
        foo = R.paste("X", [1, 2, 3])
        self.assertEqual(foo, ["X 1", "X 2", "X 3"])

        bar = R.paste("X", [1, 2, 3], sep="")
        self.assertEqual(bar, ["X1", "X2", "X3"])

        baz = R.paste([1, 2, "X"], [4, 5, 6])
        self.assertEqual(baz, ["1 4", "2 5", "X 6"])

        foo = R.paste("X", pd.Series([1, 2, 3]), sep="")
        self.assertEqual(foo, ["X1", "X2", "X3"])

        f = R.paste("X", [1, 2, 3], sep="", collapse=" ")
        self.assertEqual(f, "X1 X2 X3")


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestUtilities()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_conversions()
    rig.test_numerics()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import unittest

from BayesBoom.spikeslab import (
    dot,
    lm_spike,
    RegressionSpikeSlabPrior,
    StudentSpikeSlabPrior,
    BigAssSpikeSlab
)
from BayesBoom.R import delete_if_present

import numpy as np
import pandas as pd
import scipy.sparse
import pickle


class SpikeSlabTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def tearDown(self):
        delete_if_present("prior.pkl")

    def test_mcmc(self):
        # Run the test with a big sample size and a small residual SD, so
        # you'll be sure to find the right answer.
        sample_size = 10000
        ngood = 5
        nbad = 30
        niter = 250
        x = np.random.randn(sample_size, ngood + nbad)

        beta = np.random.randn(ngood) * 4

        b0 = 7.2
        residual_sd = .3
        yhat = b0 + x[:, :ngood] @ beta
        errors = np.random.randn(sample_size) * residual_sd
        y = yhat + errors

        data = pd.DataFrame(
            x, columns=["X" + str(i) for i in range(x.shape[1])])
        data["y"] = y
        formula = "y ~ " + dot(data, ["y"])

        model = lm_spike(formula, niter=niter, data=data)
        self.assertTrue(isinstance(
            model._coefficient_draws,
            scipy.sparse.csc_matrix)
        )
        self.assertEqual(model._coefficient_draws.shape[0], niter)
        self.assertEqual(model._coefficient_draws.shape[1], ngood + nbad + 1)

        self.assertEqual(len(model._residual_sd), niter)
        self.assertTrue(np.all(model._residual_sd > 0))

        # Test the read-only properties
        self.assertEqual(len(model._log_likelihood), niter)
        self.assertEqual(model.xdim, 1 + ngood + nbad)

        # Test the predict method.
        pred = model.predict(data, burn=3)
        self.assertTrue(isinstance(pred, np.ndarray))
        self.assertEqual(len(pred.shape), 2)
        self.assertEqual(pred.shape[0], niter-3)
        self.assertEqual(pred.shape[1], sample_size)

    def test_dot(self):
        X = pd.DataFrame(np.random.randn(10, 3), columns=["X1", "X2", "X3"])

        # Test with an empty omit.
        formula = dot(X)
        self.assertEqual(formula, "(X1+X2+X3)")

        X["y"] = 7.0
        formula = "y ~ " + dot(X, omit=["y"])
        self.assertEqual(formula, "y ~ (X1+X2+X3)")

        # Test that omit still works as a string.
        formula = "y ~ " + dot(X, omit="y")
        self.assertEqual(formula, "y ~ (X1+X2+X3)")

        # Test with multiple omitted columns.
        formula = "y ~ " + dot(X, omit=["y", "X2"])
        self.assertEqual(formula, "y ~ (X1+X3)")

    def test_regression_spike_slab_prior(self):
        n = 10
        X = pd.DataFrame(np.random.randn(n, 3), columns=["X1", "X2", "X3"])
        y = np.random.randn(n)
        prior = RegressionSpikeSlabPrior(X.values, y)
        self.assertIsInstance(prior, RegressionSpikeSlabPrior)

        fname = "prior.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(prior, pkl)

        with open(fname, "rb") as pkl:
            p2 = pickle.load(pkl)

        self.assertEqual(set(prior.__dict__.keys()),
                         set(p2.__dict__.keys()))

    def test_student_spike_slab_prior(self):
        n = 10
        X = pd.DataFrame(np.random.randn(n, 3), columns=["X1", "X2", "X3"])
        y = np.random.randn(n)
        prior = StudentSpikeSlabPrior(X.values, y)

        fname = "prior.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(prior, pkl)

        with open(fname, "rb") as pkl:
            p2 = pickle.load(pkl)

        self.assertEqual(set(prior.__dict__.keys()),
                         set(p2.__dict__.keys()))


class BigAssSpikeSlabTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_mcmc(self):
        nobs = 100000
        dim = 1000
        X = np.random.randn(nobs, dim)
        X[:, 0] = 1.0
        coefficients = np.zeros(dim)
        coefficients[0] = 28
        coefficients[3] = -72
        coefficients[84] = 54
        coefficients[93] = 180

        residual_sd = .3
        yhat = X @ coefficients
        y = yhat + np.random.randn(nobs) * residual_sd

        model = BigAssSpikeSlab(dim, subordinate_model_max_dim=50)
        i = 0
        chunk_size = 1000
        while (i < nobs):
            chunk = range(i, min(i + chunk_size, nobs))
            model.stream_data_for_initial_screen(X[chunk, :], y[chunk])
            i += chunk_size
        model.initial_screen()


_debug_mode = True

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = BigAssSpikeSlabTest()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_mcmc()

    print("Goodbye, cruel world!")

if __name__ == "__main__":
    unittest.main()



import unittest
import BayesBoom.R as R
import BayesBoom.boom as boom
import numpy as np
import pandas as pd


class TestUtilities(unittest.TestCase):

    def setUp(self):
        pass

    def test_conversions(self):
        x = [1, 2, 3]
        v = R.to_boom_vector(x)
        self.assertIsInstance(v, boom.Vector)

        x = pd.Series(x, dtype="int")
        v = R.to_boom_vector(x)
        self.assertIsInstance(v, boom.Vector)

    def test_numerics(self):
        numeric_df = pd.DataFrame(np.random.randn(10, 3))
        self.assertTrue(R.is_all_numeric(numeric_df))

        non_numeric = numeric_df.copy()
        non_numeric["text"] = "foo"
        self.assertFalse(R.is_all_numeric(non_numeric))


_debug_mode = True

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestUtilities()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_conversions()
    rig.test_numerics()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import unittest
import numpy as np
import pandas as pd
import pickle
import json

import pdb

import matplotlib.pyplot as plt

from BayesBoom.R import delete_if_present

from BayesBoom.bsts import (
    Bsts,
    AirPassengers,
    DynamicRegressionStateModel,
    LocalLevelStateModel,
    LocalLinearTrendStateModel,
    SeasonalStateModel,
)


def default_colnames(n):
    return ["X" + str(i + 1) for i in range(n)]


class TestDynamicRegression(unittest.TestCase):
    def setUp(self):
        n = 100
        xdim = 3
        coefficients = np.cumsum(np.random.randn(n, xdim) * .1, axis=0)
        self._predictors = np.random.randn(n, xdim)
        self._random_walk = np.cumsum(np.random.randn(n) * .1)
        self._reg = np.sum(coefficients * self._predictors, axis=1)
        self._noise = np.random.randn(n)
        self._y = self._random_walk + self._reg + self._noise

    def tearDown(self):
        pass
        # delete_if_present("bsts_llt.pkl")

    def test_local_level(self):
        model = Bsts()
        model.add_state(LocalLevelStateModel(self._y))
        data = pd.DataFrame(
            self._predictors,
            columns=default_colnames(self._predictors.shape[1]))
        data["y"] = self._y

        model.add_state(DynamicRegressionStateModel(
            "y ~ X1 + X2 + X3", data=data))

        model.train(data=data["y"], niter=1000)

    def test_plots(self):
        pass

    def test_predictions(self):
        pass

    def test_serialization(self):
        pass


_debug_mode = False


if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestDynamicRegression()
    # rig = TestGaussianTimeSeries()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_local_level()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



# This module contains low-level utilities for things like type-checking and
# coercion.

import numpy as np
import pandas as pd
import BayesBoom.boom as boom
from numbers import Number


def is_all_numeric(data_frame):
    """
    Returns True iff 'data_frame' is a pd.DataFrame (or equivlalent)
    """
    return np.all(data_frame.dtypes.apply(pd.api.types.is_numeric_dtype))


def is_iterable(obj):
    """
    Check to see if 'object' is an iterable object.  An iterable object allows
    the idiom 'for x in object:...'.
    """
    return hasattr(obj, "__getitem__") or hasattr(obj, "__iter__")


def to_boom_vector(v):
    """
    Convert the vector-like object 'v' to a boom.Vector.  This is a more user
    friendly experience than relying on the boom.Vector constructor, which only
    accepts floating point numpy arrays.  Here 'v' can be a numeric scalar, a
    numpy array of any numeric dtype, a pandas Series of any numeric dtype, or
    any similar object that either acts like a pd.Series or is convertible to a
    np.array.
    """
    if hasattr(v, "values"):
        # Handle pd.Series and similar.
        return boom.Vector(np.array(v.values, dtype="float"))

    if isinstance(v, Number):
        return boom.Vector(np.array([v], dtype="float"))

    return boom.Vector(np.array(v, dtype="float"))


def to_boom_matrix(m):
    """
    Convert the matrix-like object 'm' to a boom.Matrix.  This is a more user
    friendly experience than relying on the boom.Matrix constructor, which only
    accepts floating point numpy arrays.  Here 'm' can be a numeric scalar, a
    numpy array of any numeric dtype, a pandas DataFrame containing numeric
    data, or any similar object that either acts like a pd.DataFrame or is
    convertible to a np.array.
    """
    if hasattr(m, "values") and hasattr(m, "dtypes") and is_all_numeric(m):
        # Handle pd.DataFrame and similar.
        return boom.Matrix(m.values.astype("float"))

    if isinstance(m, Number):
        return boom.Matrix(np.full((1, 1), m, dtype="float"))

    return boom.Matrix(np.array(m, dtype="float"))


def to_boom_spd(m):
    """
    Convert the matrix-like object 'm' to a boom.Matrix.  This is a more user
    friendly experience than relying on the boom.Matrix constructor, which only
    accepts floating point numpy arrays.  Here 'm' can be a numeric scalar, a
    numpy array of any numeric dtype, a pandas DataFrame containing numeric
    data, or any similar object that either acts like a pd.DataFrame or is
    convertible to a np.array.
    """
    if hasattr(m, "values") and hasattr(m, "dtypes") and is_all_numeric(m):
        # Handle pd.DataFrame and similar.
        return boom.SpdMatrix(m.values.astype("float"))

    if isinstance(m, Number):
        return boom.SpdMatrix(np.full((1, 1), m, dtype="float"))

    return boom.SpdMatrix(np.array(m, dtype="float"))


def to_boom_date(timestamp):
    """
    Convert a pd.Timestamp or similar to a boom.Date object.
    """
    if isinstance(timestamp, str):
        timestamp = pd.Timestamp(timestamp)

    return boom._date(int(timestamp.month), int(timestamp.day),
                      int(timestamp.year))



import numpy as np


def covers(draws, truth, confidence):
    """
    Private function used to implement check_mcmc_vector and check_mcmc_matrix.
    Returns True iff a centeral 'confidence' interval constructed from 'draws'
    contains 'truth'.

    Args:
      draws:  A numpy vector containing Monte Carlo draws of a quantity.
      truth:  The target value that 'draws' are trying to cover.
     confidence: The probability content of the credibility interval used to
       check coverage.
    """
    sorted_draws = np.sort(np.array(draws).flatten())
    lower = .5 * (1 - confidence)
    upper = 1 - lower
    n = sorted_draws.shape[0]
    lo = sorted_draws[int(np.round(n * lower))]
    hi = sorted_draws[int(np.round(n * upper))]
    return truth >= lo and truth <= hi


def check_mcmc_vector(draws, truth, confidence=0.95):
    """
    Check to see if a vector of Monte Carlo draws covers a known value.

   Args:
     draws:  The vector of Monte Carlo draws to check.
     truth:  The true value against which 'draws' will be checked.
     confidence: The probability content of the credibility interval used to
       check coverage.

   Returns:

     A boolean.  True indicates that a central credibility interval with
     probability content 'confidence', constructed from 'draws', covers
     'truth'.  A False value indicates that the interval does not cover.
    """
    return covers(draws, truth, confidence)


def check_mcmc_matrix(draws, truth, confidence: float = .95,
                      control_multiple_comparisons: bool = True):
    """
    Args:
      draws: numpy matrix of draws.  Each row is a draw.
      truth: numpy vector of true values, with length matching the number of
        columns in draws.
     confidence: The probability content of the credibility interval used to
       check coverage.
     control_multiple_comparisons: If True, then the fraction of intervals
       successfully covering the true value can be slightly less than
       'confidence' as long as it is consistent with the null hypothesis of
       'confidence'-level coverage, in the sense that it is no less than two
       binomial standard errors below 'confidence'.
    """
    if confidence <= 0 or confidence >= 1:
        raise Exception("Confidence must be between 0 and 1.")
    if confidence < .5:
        confidence = 1 - confidence

    draws = np.array(draws)
    if len(draws.shape) != 2:
        raise Exception("draws must be a matrix")

    dim = draws.shape[1]
    if len(truth) != dim:
        raise Exception("The vector of true values must match the "
                        "number of columns in draws.")

    coverage_indicators = np.full(dim, False)
    fails_to_cover = 0
    for i in range(dim):
        coverage_indicators[i] = covers(draws[:, i], truth[i], confidence)
        if not coverage_indicators[i]:
            fails_to_cover += 1

    fraction_failing_to_cover = fails_to_cover / dim
    coverage_rate_limit = confidence
    if control_multiple_comparisons:
        se = np.sqrt(confidence * (1 - confidence) / dim)
        coverage_rate_limit -= 2 * se
    failure_rate_limit = 1 - coverage_rate_limit
    return fraction_failing_to_cover < failure_rate_limit


def check_stochastic_process(draws: np.ndarray,
                             truth: np.ndarray,
                             confidence: float = .95,
                             sd_ratio_threshold: float = .1,
                             control_multiple_comparisons: bool = True):
    """
    Args:
      draws: A matrix of Monte Carlo draws to be checked.  Each row is a draw
        and each column is a variable.
      truth: A vector of true values against which draws will be compared.
        truth.size() must match ncol(draws).
      confidence: The confidence associated with the marginal posterior
        intervals used to determine coverage.
      sd_ratio_threshold: One of the testing diagnostics compares the standard
        deviation of the centered draws to the standard deviation of the true
        function.  If that ratio is less than this threshold the diagnostic is
        passed.

    Returns:
      A string containing an error message describing the mode of the failure
      to cover.
    """
    import BayesBoom.boom as boom
    return boom.check_stochastic_process(
        boom.Matrix(draws),
        boom.Vector(truth),
        float(confidence),
        float(sd_ratio_threshold),
        bool(control_multiple_comparisons))



import numpy as np


def suggest_burn(loglike, fraction: float = 0.1, quantile: float = 0.9):
    """Suggest a burn-in sample size for an MCMC.  The suggestion uses a heuristic
    based on the time series of log likelihood values by the MCMC sampler.  The
    algorithm looks at the last 'fraction' of the log likelihood sequence and
    finds a specified quantile to use as a threshold.  The burn-in period ends
    with the first element in the chain to meet or exceed this threshold.

    :param log.likelihood:
        MCMC sample path of log likelihood.

    :param fraction:
        The final fraction of the chain available for to determining the log
        likelihood threshold.

    :param quantile:
        The quantile of the values in the final fraction used to determine the
        threshold.

    :return burn: int
        The number of burn-in iterations that should be discarded.  This can be
        0 if 'fraction' is zero, which signals that no burn-in is desired.

    """
    if fraction < 0:
        # A signal that no burnin is desired.
        return 0
    if fraction > 1.0:
        raise Exception("'fraction' must be a number between 0 and 1.")
    cutpoint = int((1.0 - fraction) * len(loglike))
    if cutpoint == 0:
        return 0
    min_loglike = np.quantile(loglike[cutpoint:], quantile)
    burn = np.where(loglike >= min_loglike)[0][0]
    if burn < 0:
        burn = 0
    return int(burn)



from abc import ABC, abstractmethod
import BayesBoom.R as R
import numpy as np


# ===========================================================================
class StateModel(ABC):
    """StateModel objects are wrappers around boom.StateModel objects, which are
    opaquie C++ types.  The primary purpose of a separate python object is to
    handle various accounting duties required of StateModel's by the Bsts model
    defined below.

    """

    def __init__(self):
        """
        Args:
          state_index_begin: The index of the state subcomponent described by
            this state model, in the full state vector.

        """
        self._state_index = -1

    @property
    def state_index(self):
        """
        The index of the this state model's first element, in the global state
        vector.  This starts off as -1.  When the state model is added to a
        bsts model, bsts will call set_state_index() to inform the state model
        of this value.
        """
        if self._state_index < 0:
            raise Exception("Each state model must be told where its state"
                            "component begins in the global state vector.  "
                            "Try calling set_state_index.")
        return self._state_index

    def set_state_index(self, state_index_begin: int):
        """
        Inform the state model of the first index in the larger state vector
        corresponding to the state component it is responsible for.

        Args:
           state_index_begin: The smallest index in the global state vector
             corresponding to the component of state that this model describes.
        """
        self._state_index = state_index_begin

    @property
    @abstractmethod
    def label(self):
        """
        A string indicating how this state model should be labelled, e.g. when
        assessing contributions of mutliple state components.
        """

    @property
    @abstractmethod
    def state_dimension(self):
        """
        The dimension of the state subcomponent managed by this model.
        """

    @abstractmethod
    def allocate_space(self, niter, time_dimension):
        """
        Allocate the space needed to call 'record_state' the given number of
        times.

        Args:
          niter:  Number of iterations (draws) to be stored.
          time_dimension:  Number of time points in the training data.
        """

    @abstractmethod
    def record_state(self, iteration: int, state_matrix: np.ndarray):
        """
        Record the state of any model parameters, and the subset of the state
        vector associated with this model, so they can be analyzed later.

        Args:
          iteration:  The (integer) index of the MCMC iteration to be recorded.
          state_matrix: The current matrix containing state.  Each column is a
            state vector associated with the corresponding time point.

        Effect:
          The current state of the model parameters is stored.
        """

    @abstractmethod
    def restore_state(self, iteration: int):
        """
        Restore the state of the managed boom state model to the requested
        iteration.
        """

    @property
    @abstractmethod
    def state_contribution(self):
        """
        A niter x time_dimension array giving the contribution of this state
        component to the overall mean of y at each point in time.
        """

    def plot_state_contribution(
            self, fig, gridspec, time, burn=None, ylim=None, **kwargs):
        self.plot_state_contribution_default(
            fig=fig, gridspec=gridspec, time=time, burn=burn, ylim=ylim,
            **kwargs)

    def plot_state_contribution_default(
            self, fig, gridspec, time, burn=None, ylim=None, **kwargs):
        """
        A default implementation of plot_state_distribution.

        Plots the contribution of this state model to the conditional mean at
        time t on the supplied set of axes using a dynamic distribution plot.

        Args:
          ax:  The axes object on which to draw the plot.
          time:  The time axis.
          burn:  The number of iterations to discard as burn-in.
          ylim:  A pair giving the lower and upper limits on the y axis.
          kwargs:  Extra arguments passed to plot_dynamic_distribution.

        Effects:
          Creates a dynamic distribution plot on the supplied set of axes.

        Returns:
          The axes object.
        """
        if burn > 0:
            curves = self._state_contribution[int(burn):, :]
        else:
            curves = self._state_contribution

        ax = fig.add_subplot(gridspec)

        R.plot_dynamic_distribution(
            curves=curves,
            timestamps=time,
            ax=ax,
            ylim=ylim,
            **kwargs)

        return ax

    def observe_time_dimension(self, time_dimension):
        """
        Args:
          time_dimension: The number of time points being modeled.  For
            training, this is the length of the training data.  For prediction
            it is the length of the training data plus the forecast horizon.
        """
        self._state_model.observe_time_dimension(time_dimension)



import numpy as np
import BayesBoom.boom as boom
import BayesBoom.R as R
from numbers import Number
from .state_models import StateModel


class ArStateModel(StateModel):
    """
    Latent state described by a stationary AR(p) process.
    """

    def __init__(self,
                 y,
                 lags: int = 1,
                 sigma_prior: R.SdPrior = None,
                 initial_state_prior: R.MvnPrior = None,
                 sdy: float = None):
        """
        Args:
          y:  The time series to be modeled.
          lags:  The number of lags in the autoregressive process.
          sigma_prior: The prior distribution for the standard deviation
            parameter in the AR1 process.
          initial_state_prior: Prior distribution for the value of the state at
            time 0.
          sdy: The standard deviation of the time series to be modeled.  This
            argument is ignored if y is provided.
        """
        self._lags = lags
        self._state_contribution = None
        if sdy is None:
            sdy = np.nanstd(y)

        self._validate_prior(sigma_prior, sdy)
        self._validate_initial_distribution(initial_state_prior, sdy)
        self._build_state_model()

    @property
    def label(self):
        return f"AR({self._lags})"

    @property
    def lags(self):
        return self._lags

    @property
    def state_dimension(self):
        return self._lags

    @property
    def state_error_dimension(self):
        return 1

    @property
    def state_contribution(self):
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self._sigma = np.empty(niter)
        self._ar_coefficients = np.empty((niter, self.state_dimension))
        self._state_contribution = np.empty((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self._sigma[iteration] = self._state_model.sigma
        self._ar_coefficients[iteration, :] = self._state_model.phi.to_numpy()
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

    def restore_state(self, iteration):
        self._state_model.set_sigma(self._sigma[iteration])
        self._state_model.set_phi(R.to_boom_vector(
            self._ar_coefficients[iteration, :]))

    def _validate_prior(self, prior, sdy):
        if prior is None:
            prior = R.SdPrior(sdy * .01, upper_limit=sdy)
        if not isinstance(prior, R.SdPrior):
            raise Exception("Wrong type for prior.")
        self._sigma_prior = prior

    def _validate_initial_distribution(self, initial_state_prior, sdy):
        if initial_state_prior is None:
            dim = self.state_dimension
            initial_state_prior = R.MvnPrior(np.zeros(dim), np.eye(dim) * sdy)
        if not isinstance(initial_state_prior, R.MvnPrior):
            raise Exception("Wrong type for initial_state_prior.")
        if initial_state_prior.dim != self.state_dimension:
            raise Exception(
                f"Initial_state_prior dimension was {initial_state_prior.dim}."
                f"  State dimension is {self.state_dimension}."
            )
        self._initial_state_prior = initial_state_prior

    def _build_state_model(self):
        self._state_model = boom.ArStateModel(self._lags)
        self._set_posterior_sampler()
        self._set_initial_distribution()

    def _set_posterior_sampler(self):
        sampler = boom.ArPosteriorSampler(
            self._state_model,
            self._sigma_prior.boom(),
            boom.GlobalRng.rng)
        limit = self._sigma_prior.upper_limit
        if np.isfinite(limit) and limit > 0:
            sampler.set_sigma_upper_limit(limit)
        self._state_model.set_method(sampler)

    def _set_initial_distribution(self):
        self._state_model.set_initial_state_mean(
            R.to_boom_vector(self._initial_state_prior.mean))
        self._state_model.set_initial_state_variance(
            R.to_boom_spd(self._initial_state_prior.variance))

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()


class SpikeSlabArPrior:
    def __init__(self,
                 lags,
                 sdy=None,
                 prior_inclusion_probabilities=None,
                 prior_mean=None,
                 prior_sd=None,
                 prior_df=1,
                 expected_r2=0.5,
                 sigma_upper_limit=np.Inf,
                 max_flips=np.Inf):
        self._lags = int(lags)
        if self._lags < 0:
            raise Exception("'lags' argument must be non-negative.")

        self._max_flips = max_flips

        if prior_inclusion_probabilities is None:
            prior_inclusion_probabilities = np.geomspace(
                .8, .8 ** self._lags, num=self._lags)
        self._prior_inclusion_probabilities = np.array(
            prior_inclusion_probabilities)

        if prior_mean is None:
            self._prior_mean = np.zeros(self._lags)
        else:
            self._prior_mean = np.array(prior_mean, dtype=float)
        if len(self._prior_mean) != self._lags:
            raise Exception("prior_mean argument must have length 'lags'.")

        if prior_sd is None:
            self._prior_sd = np.geomspace(
                .5, .5 * .8 ** (self._lags - 1), num=self._lags)
        elif isinstance(self._prior, Number):
            self._prior_sd = np.array([prior_sd] * self._lags)
        else:
            self._prior_sd = np.array(prior_sd, dtype=float)

        sigsq_guess = sdy * expected_r2
        self._residual_precision = R.SdPrior(
            sigma_guess=np.sqrt(sigsq_guess),
            sample_size=prior_df,
            upper_limit=sigma_upper_limit)

    def __repr__(self):
        f"""
        SpikeSlabArPrior with...
        inclusion_probabilities: {self._prior_inclusion_probabilities}
        max flips: {self._max_flips}
        prior mean: {self._prior_mean}
        prior sd: {self._prior_sd}
        residual precision prior: {self._residual_precision}
        """

    @property
    def spike(self):
        return boom.VariableSelectionPrior(
            R.to_boom_vector(
                self._prior_inclusion_probabilities))

    @property
    def slab(self):
        return boom.MvnModel(
            R.to_boom_vector(self._prior_mean),
            R.to_boom_spd(np.diag(self._prior_sd ** 2))
        )

    @property
    def residual_precision(self):
        return self._residual_precision

    @property
    def max_flips(self):
        return self._max_flips


class AutoArStateModel(ArStateModel):
    def __init__(self,
                 y,
                 lags: int = 1,
                 prior: SpikeSlabArPrior = None,
                 initial_state_prior: R.MvnPrior = None,
                 sdy: float = None,
                 **kwargs):
        if sdy is None:
            sdy = np.nanstd(y)
        super().__init__(y, lags, prior, initial_state_prior, sdy)

    @property
    def label(self):
        return f"AutoAr({self.lags})"

    def _validate_prior(self, prior, sdy):
        if prior is None:
            prior = SpikeSlabArPrior(
                lags=self.lags,
                sdy=sdy,
                sigma_upper_limit=sdy)

        if not isinstance(prior, SpikeSlabArPrior):
            raise Exception("Wrong type for 'prior'.")
        self._prior = prior

    def _set_posterior_sampler(self):
        sampler = boom.ArSpikeSlabSampler(
            self._state_model,
            self._prior.slab,
            self._prior.spike,
            self._prior.residual_precision.boom(),
            True,
            boom.GlobalRng.rng)
        limit = self._prior.residual_precision.upper_limit
        if limit > 0 and np.isfinite(limit):
            sampler.set_sigma_upper_limit(limit)
        max_flips = self._prior.max_flips
        if max_flips > 0 and np.isfinite(max_flips):
            sampler.limit_model_selection(max_flips)

        self._state_model.set_method(sampler)

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()



import unittest
import pandas as pd

import pdb

from BayesBoom.bsts import (
    # Bsts,
    # RegressionHolidayStateModel,
    FixedDateHoliday,
    LastWeekdayInMonthHoliday,
    # NthWeekdayInMonthHoliday,
    EasterSunday,
    USDaylightSavingsTimeBegins,
    USDaylightSavingsTimeEnds,
)


class TestHolidays(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass
        # delete_if_present("bsts_llt.pkl")

    def test_boom_holidays(self):
        easter = EasterSunday()
        self.assertAlmostEqual(
            easter.date(2021),
            pd.Timestamp(year=2021, month=4, day=4))
        self.assertAlmostEqual(
            easter.date(2020),
            pd.Timestamp(year=2020, month=4, day=12))

        dst_begin = USDaylightSavingsTimeBegins()
        self.assertAlmostEqual(
            dst_begin.date(2021),
            pd.Timestamp(year=2021, month=3, day=14))

        dst_end = USDaylightSavingsTimeEnds()
        self.assertAlmostEqual(
            dst_end.date(2020),
            pd.Timestamp(year=2020, month=11, day=1))

        july4 = FixedDateHoliday(month=7, day=4)
        self.assertAlmostEqual(
            july4.date(2020),
            pd.Timestamp(year=2020, month=7, day=4))

        memorial_day = LastWeekdayInMonthHoliday(1, 5)
        self.assertAlmostEqual(
            memorial_day.date(2020),
            pd.Timestamp(year=2020, month=5, day=25))


class TestRegressionHolidayStateModel(unittest.TestCase):
    pass


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestHolidays()
    # rig = TestGaussianTimeSeries()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_boom_holidays()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import scipy.stats as sp
import numpy as np

# expose R's equivalent of pnorm, qnorm, etc.


def as_numeric(x):
    return np.array(x)


def dnorm(x, mean=0, sd=1, log=False):
    """Normal density function.

    :param x: number, list or numpy array
        The argument of the density function.

    :param x: number, list or numpy array
        The argument of the density function.

    """
    x = as_numeric(x)
    if log:
        return sp.norm.logpdf(x, mean, sd)
    else:
        return sp.norm.pdf(x, mean, sd)


def pnorm(x, mean=0, sd=1, lower=True, log=False):
    """
    Normal cumulative distribuion function.

    """
    if log:
        if lower:
            return sp.norm.logcdf(x, loc=mean, scale=sd)
        else:
            return sp.norm.logsf(x, loc=mean, scale=sd)
    else:
        if lower:
            return sp.norm.cdf(x, loc=mean, scale=sd)
        else:
            return sp.norm.sf(x, loc=mean, scale=sd)


def qnorm(x, mean=0, sd=1, lower=True, log=False):
    """
    Quantiles of the normal distribuion.

    """
    if log:
        x = np.exp(x)

    if lower:
        return sp.norm.ppf(x, mean, sd)
    else:
        return sp.norm.ppf(1 - x, mean, sd)


def rnorm(n, mean=0, sd=1):
    """
    Random deviates from the normal distribution.

    """

    return np.random.randn(n) * sd + mean


def dgamma(y, shape, scale, log=False):
    from scipy.stats import gamma
    if log:
        return gamma.logpdf(y, shape, scale=1.0 / scale)
    else:
        return gamma.pdf(y, shape, scale=1.0 / scale)


def pgamma(y, shape, scale, lower_tail=True, log=False):
    from scipy.stats import gamma
    if lower_tail:
        if log:
            return gamma.logcdf(y, shape, scale=1.0 / scale)
        else:
            return gamma.cdf(y, shape, scale=1.0 / scale)
    else:
        if log:
            return gamma.logsf(y, shape, scale=1.0 / scale)
        else:
            return gamma.sf(y, shape, scale=1.0 / scale)


def qgamma(probs, shape, scale, lower_tail=True, log=False):
    from scipy.stats import gamma
    if log:
        probs = np.exp(probs)
    if lower_tail:
        return gamma.ppf(probs, shape, scale=1.0 / scale)
    else:
        return gamma.isf(probs, shape, scale=1.0 / scale)


def rgamma(n, shape, scale):
    """
    Random deviates from the gamma distribution with mean shape / scale.
    """
    from scipy.stats import gamma
    return gamma.rvs(shape, size=n, scale=1.0 / scale)


def rmarkov(n: int, P: np.ndarray, pi0=None):
    """
    Simulation a Markov chain of length n from transition probability matrix P,
    with initial distribution pi0.  The state space is the set of integers in
    {0, ..., S-1} where S is the number of rows in P.

    Args:
      n:  The length of the Markov chain to simulate.
      P: A transition probability matrix.  A square matrix of non-negative
         elements, where each row sums to 1.
      pi0: The distribution of the state at time 0.  A vector of non-negative
        numbers summing to 1.

    Returns:
      A numpy array of integer dtype containing the simulated Markov chain..
    """
    P = np.array(P)
    if len(P.shape) != 2 or P.shape[0] != P.shape[1]:
        raise Exception("P must be a square matrix.")
    S = P.shape[0]
    if pi0 is None:
        pi0 = np.ones(S) / S
    else:
        pi0 = np.array(pi0).flatten()

    if len(pi0) != S:
        raise Exception("Initial distribution must have the same "
                        "dimension as the transition probability matrix.")
    ans = np.full(n, -1)
    ans[0] = np.random.choice(range(S), p=pi0)
    for i in range(1, n):
        ans[i] = np.random.choice(range(S), p=P[ans[i-1], :])
    return(ans)



from .bsts import ObservationModelManager
import patsy
import numpy as np
import BayesBoom.boom as boom
import BayesBoom.R as R
import BayesBoom.spikeslab as spikeslab
import scipy


class StateSpaceStudentModelFactory:
    def __init__(self, formula: str):
        """
        Create a StateSpaceStudentModelFactory.

        Args:
          formula: A string describing how the predictor variables are to be
            constructed from a data frame.  For example "y ~ x1 + x2 + x2*x3".
            If the formula is None then the model will contain no regression
            component.
        """
        if formula is not None and not isinstance(formula, str):
            raise Exception("formula must either be None or a string")
        self._formula = formula
        self.predictor_names = None

    def create_model(self, prior, data, rng, **kwargs):
        """
        Create the boom model object, and store related model artifacts.

        Args:
          prior: The prior for the observation model.  Either None or a
            spikeslab.StudentSpikeSlabPrior.
          data: If self._formula is None then data is the time series to be
            modeled.  Either a pd.Series or a np.ndarray.  Otherwise data
            should be a pd.DataFrame containing the variables referenced in
            'formula'.
          **kwargs: If prior is None then any remaining arguments are passed to
            the StudentSpikeSlabPrior constructor.

        Returns:
          The created model.

        Effects:
          self._model is populated with the created model.
          self._prior is populated with the prior distribution for the
            observation model.
        """

        if data is not None:
            if self._formula is None:
                # Pure time series case.
                response = data
                predictors = np.ones((len(response), 1))
                kwargs["expected_model_size"] = 0
            else:
                # Time series regression case.
                response, predictors = patsy.dmatrices(self._formula, data)
                self.predictor_names = predictors.design_info.term_names

            boom_response = boom.Vector(R.to_numpy(response))
            boom_predictors = boom.Matrix(R.to_numpy(predictors))
            response_is_observed = np.isfinite(response).ravel()

            self._model = boom.StateSpaceStudentRegressionModel(
                boom_response, boom_predictors, response_is_observed)
        elif prior is not None:
            xdim = len(prior._prior_inclusion_probabilities)
            self._model = boom.StateSpaceStudentRegressionModel(xdim)
            response = None
            predictors = None
        else:
            raise Exception("At least one of 'data' or 'prior' is needed.")

        regression = self._model.observation_model

        prior = self._verify_prior(prior, response, predictors, **kwargs)

        self._prior = prior

        observation_model_sampler = boom.TRegressionSpikeSlabSampler(
            regression,
            prior.slab(regression.Sigsq_prm),
            prior.spike,
            prior.residual_precision,
            prior.tail_thickness,
            rng,
        )
        observation_model_sampler.set_sigma_upper_limit(prior.sigma_upper_limit)
        if prior.max_flips > 0:
            observation_model_sampler.limit_model_selection(prior.max_flips)
        regression.set_method(observation_model_sampler)

        sampler = boom.StateSpaceStudentPosteriorSampler(
            self._model,
            observation_model_sampler)
        self._model.set_method(sampler)

        self._original_series = response
        return self._model

    def create_observation_model_manager(self):
        return StudentObservationModelManager(
            self._model.observation_model.xdim,
            self._formula)

    def _verify_prior(self, prior, response, predictors, **kwargs):
        """
        Ensure that the prior is set up right.
        Args:
          prior: Either None, or a spikeslab.StudentSpikeSlabPrior object.  If
            None then a default prior will be created from the supplied data.
          response: If prior is None, this must either be a vector containing
            the response variable.  Acceptable types include a boom.Vector,
            pd.Series, or a np.ndarray.  If prior is None then this value is
            not used.
          predictors: If prior is None, this must be a Matrix containing the
            predictor values.  Acceptable types include a boom.Matrix,
            pd.DataFrame with all numeric dtypes, or a np.ndarray.
          **kwargs: If prior is None then any additional named arguments are
            passed to the StudentSpikeSlabPrior constructor.

        Returns:
          The validated prior.  This is 'prior' if not None and of the correct
          type.  Otherwise it is the default prior constructed from the
          remaining arguments.

        Exceptions:
          An exception is thrown if 'prior' is not None and it is not a
          StudentSpikeSlabPrior.
        """
        if prior is None:
            prior = spikeslab.StudentSpikeSlabPrior(
                predictors,
                response,
                **kwargs,
            )

        if not isinstance(prior, spikeslab.StudentSpikeSlabPrior):
            raise Exception("Expected a StudentSpikeSlabPrior prior "
                            "distribution.")
        return prior


class StudentObservationModelManager(ObservationModelManager):
    def __init__(self, xdim: int, formula: str):
        """
        In a pure time series context, xdim == 0.  In a regression context,
        xdim is the number of predictors in the regression portion of the model.
        """
        self._xdim = xdim
        self._formula = formula

    @property
    def has_regression(self):
        return self._xdim > 1 and self._formula is not None

    def allocate_space(self, niter: int, time_dimension: int):
        self._residual_sd = np.empty(niter)
        self._residual_df = np.empty(niter)
        if self._xdim > 0:
            self._coefficients = scipy.sparse.lil_matrix((niter, self._xdim))

        if self.has_regression:
            self._regression_contribution = np.empty((niter, time_dimension))

    def record_draw(self, iteration: int, model):
        student_reg = model.observation_model
        self._residual_sd[iteration] = student_reg.residual_sd
        self._residual_df[iteration] = student_reg.residual_df
        if self._xdim > 0:
            self._coefficients[iteration, :] = spikeslab.sparsify(
                student_reg.coef)

        if self.has_regression:
            self._regression_contribution[iteration, :] = (
                model.regression_contribution.to_numpy()
            )

    def restore_draw(self, iteration: int, model):
        student_reg = model.observation_model
        student_reg.set_residual_sd(self._residual_sd[iteration])
        student_reg.set_residual_df(self._residual_df[iteration])
        if self._xdim > 0:
            spikeslab.set_glm_coefs(
                student_reg.coef,
                self._coefficients[iteration, :])

    def format_prediction_data(self, prediction_data, **kwargs):
        if isinstance(prediction_data, int):
            formatted = {
                "forecast_horizon": int(prediction_data),
                "predictors": boom.Matrix(np.ones((int(prediction_data), 1)))
            }
        else:
            predictor_matrix = patsy.dmatrix(
                self._formula, data=prediction_data)
            xnames = predictor_matrix.design_info.term_names
            formatted = {
                "forecast_horizon": prediction_data.shape[0],
                "predictors": boom.Matrix(predictor_matrix),
                "xnames": xnames,
            }
        return formatted

    def predict(self, model, formatted_prediction_data, boom_final_state, rng,
                separate_components=False, **kwargs):
        """
        Return one draw from the posterior predictive distribution.
        """
        if separate_components:
            draw = model.simulate_forecast_components(
                rng, formatted_prediction_data["predictors"], boom_final_state)
        else:
            draw = model.simulate_forecast(
                rng, formatted_prediction_data["predictors"], boom_final_state)
        return draw.to_numpy()



import numpy as np
import pandas as pd
from abc import ABC, abstractmethod


# ===========================================================================
class Encoder(ABC):
    """
    An encoder converts a data frame (which might contain mixed type data:
    numeric, categorical, datetime) into a numeric matrix of predictors.

    The most common high-level encoder is a DatasetEncoder, which does the
    conversion mentioned above in a single call: encoder.encode_dataset(data).
    A DatasetEncoder is composed of other smaller encoders that do things like
    expand variables into spline basis expansions, dummy variables, and
    interactions.
    """

    @property
    @abstractmethod
    def dim(self):
        """
        The dimension of the basis element output by this encoder.
        """

    @property
    @abstractmethod
    def encoded_variable_names(self):
        """
        The column names of the predictor matrix produced by 'encode_dataset..
        """

    @abstractmethod
    def encode_dataset(self, data):
        """
        Args:
          data:  A pd.DataFrame or equivalent containing mixed-type data.

        Returns:
          predictors:  A np.ndarray containing the encoded numeric predictors.
        """

    @property
    @abstractmethod
    def required_variables(self):
        """
        A list of the variable names needed in order to call encode_dataset.
        """

    @abstractmethod
    def encodes(self, vname):
        """
        Returns True if the encoder encodes a variable named 'vname' and False
        otherwise.
        """


# ===========================================================================
class MainEffectEncoder:
    """
    A "Main Effect" is a variable in a model that is a function of a single
    column in the input data frame.
    """

    def __init__(self, variable_name: str):
        self._vname = variable_name

    @abstractmethod
    def encode(self, x):
        """
        Args:
          x: a pd.Series or equivalent.  Depending on the type of the concrete
            encoder, x might be numeric, categorical, datetime, or other
            non-numeric data type.

        Returns:
          encoded: A np.ndarray.  Each row matches the corresponding entry in
            'x'.  The number of columns is self.dim.
        """

    def encode_dataset(self, data):
        return self.encode(data.loc[:, self.variable_name])

    @property
    def variable_name(self):
        return self._vname

    @property
    def required_variables(self):
        return [self.variable_name]

    def encodes(self, vname):
        return vname == self.variable_name

    def __repr__(self):
        return str(self.__class__.__name__) + " for " + self.variable_name


# ===========================================================================
class EffectEncoder(MainEffectEncoder):
    """
    Convert categorical data into a matrix of dummy variables.  An
    EffectEncoder is similar to a "1-hot" encoder for all levels other than the
    baseline level.

    Each column of the output matrix corresponds to one non-baseline level of a
    categorical variable.
    - If the categorical is the "special" level that column gets the value 1.
    - If the categorical variable is neither the special level nor the baseline
      level then that column gets the value 0.
    - If the categorical variable is the baseline level then the column gets
      the value -1.

    The logic behind this encoding scheme is the effect of the baseline level
    is the negative sum of the effect of all other levels.  In other words, the
    set of effects for the categorical variable is constrained to sum to zero.

    Example:

      Suppose a variable Color is either Red, Blue, or Green, with Green being
      the baseline.  If the first three instances of Color happened to be Red,
      Blue, and Green in that order, the first three elements of the encoded
      output will be:

         Color.Red  Color.Blue
         1          0            # Red
         0          1            # Blue
        -1         -1            # Green

      Now suppose that the coefficient of "Color.Red" has the value 1.2, and
      the coefficient of Color.Blue has the value 2.3.  Multiply these
      coefficient by the rows of the encoded matrix to get

      The effect of "Red" = 1.2
      The effect of "Blue" = 2.3
      The effect of "Green" = -1.2 - 2.3.

      Obviously, the sum of these three effects is zero.

    Effect encoding has the appealing property that the an as-yet unseen level
    of the categorical variable.
    """

    def __init__(self, variable_name, levels, baseline_level=None):
        super().__init__(variable_name)
        self._levels = self._unique_levels(levels)
        assert isinstance(self._levels, list)

        if baseline_level is None:
            self._baseline = levels[-1]
        else:
            self._baseline = baseline_level
        if self._baseline in self._levels:
            self._levels.remove(self._baseline)

    @property
    def dim(self):
        return len(self._levels)

    def encode(self, factor):
        if not isinstance(factor, (np.ndarray, pd.Series)):
            factor = np.array(factor)
        nobs = len(factor)
        dim = self.dim
        ans = np.zeros((nobs, dim), dtype=float)
        for i in range(dim):
            ans[:, i] = (factor == self._levels[i]).astype(float)
        baseline = factor == self._baseline
        ans[baseline, :] = -1
        return ans

    @property
    def encoded_variable_names(self):
        return [self.variable_name + "." + str(x) for x in self._levels]

    @staticmethod
    def _unique_levels(levels):
        ans = []
        used = set()
        for level in levels:
            if level not in used:
                ans.append(level)
                used.add(level)
        return ans


# ===========================================================================
class IdentityEncoder(MainEffectEncoder):
    """
    An IdentityEncoder encodes a single numeric column into and n x 1 matrix.
    """

    def __init__(self, variable_name):
        super().__init__(variable_name)

    @property
    def dim(self):
        return 1

    def encode(self, x):
        return np.array(x).reshape(-1, 1)

    @property
    def encoded_variable_names(self):
        return self.variable_name


# ===========================================================================
class InteractionEncoder(Encoder):
    """
    Creates an interaction between two lower dimensional encoders.  Multi-way
    interactions can be implemented in terms of lower-order interactions.
    """

    def __init__(self, encoder1: Encoder, encoder2: Encoder):
        self._encoder1 = encoder1
        self._encoder2 = encoder2
        self._dim = self._encoder1.dim * self._encoder2.dim

    @property
    def dim(self):
        return self._encoder1.dim * self._encoder2.dim

    def encoded_variable_names(self):
        names1 = self._encoder1.encoded_variable_names
        names2 = self._encoder2.encoded_variable_names
        ans = []
        for name1 in names1:
            for name2 in names2:
                ans.append(name1 + ":" + name2)
        return ans

    def encode(self, factor1, factor2):
        nobs = len(factor1)
        if len(factor2) != nobs:
            raise Exception("Both factors in an interaction must have the "
                            "same number of observations.")

        enc1 = self._encoder1.encode(factor1)
        enc2 = self._encoder2.encode(factor2)
        return self.create_interaction(enc1, enc2)

    def create_interaction(self, basis1, basis2):
        """
        Returns a matrix that contains all column-wise products between the
        elements of basis1 and basis2, which are matrices with the same number
        of rows.
        """
        dim = basis1.shape[1] * basis2.shape[1]
        nobs = basis1.shape[0]
        if basis2.shape[0] != nobs:
            raise Exception("Basis matrices must have the same number of rows")
        ans = np.empty((nobs, dim))
        counter = 0
        for col1 in range(basis1.shape[1]):
            for col2 in range(basis2.shape[1]):
                ans[:, counter] = basis1[:, col1] * basis2[:, col2]
                counter += 1
        return ans

    def encode_dataset(self, data):
        return self.create_interaction(
            self._encoder1.encode_dataset(data),
            self._encoder2.encode_dataset(data))

    @property
    def required_variables(self):
        return list(set(self._encoder1.required_variables
                        + self._encoder2.required_variables))

    def encodes(self, vname):
        return self._encoder1.encodes(vname) or self._encoder2.encodes(vname)


# ===========================================================================
class DatasetEncoder(Encoder):

    def __init__(self, encoders, force_intercept: bool = True):
        self._encoders = encoders
        self._force_intercept = force_intercept

        self._initialize()

    @property
    def dim(self):
        return self._dim

    @property
    def encoded_variable_names(self):
        ans = []
        if self._force_intercept:
            ans.append("(Intercept)")
        for enc in self._encoders:
            ans += enc.encoded_variable_names
        return ans

    def encode_dataset(self, data: pd.DataFrame):
        """
        Encode the data set.

        Args:
          data: The data to be encoded.

        Returns:
          The matrix of encoded data (as an np.ndarray)
        """
        if data.shape[0] == 0:
            return np.empty((0, self.dim))
        encoded_variables = []
        if self._force_intercept:
            encoded_variables.append(np.ones((data.shape[0], 1)))
        for i in range(len(self._encoders)):
            encoded_variables.append(self._encoders[i].encode_dataset(data))
        if not encoded_variables:
            return np.array([])
        else:
            return np.concatenate(encoded_variables, axis=1)

    def _initialize(self):
        self._dim = self._force_intercept + np.sum(
            [x.dim for x in self._encoders])
        self._factor_map = self._create_variable_map()

    @property
    def required_variables(self):
        required = []
        for enc in self._encoders:
            required += enc.required_variables
        return list(set(required))

    def encodes(self, vname):
        for enc in self._encoders:
            if enc.encodes(vname):
                return True
        return False

    def _create_variable_map(self):
        vnames = self.required_variables
        affected_feature_indices = {}
        main_effect_indices = {}
        for name in vnames:
            affected, main = self._affected_features(name)
            affected_feature_indices[name] = affected
            main_effect_indices[name] = main
        return VariableMap(affected_feature_indices, main_effect_indices)

    def _affected_features(self, vname: str):
        """
        The set of output columns affected by the named variable.

        Args:
          vname:  The name of an input variable.

        Returns:
        - affected_indices: A list of column numbers affected by the named
          variable.
        - main_effect_indices: A list of column numbers affected by just the
          main effect of the named variable.
        """
        start = int(self._force_intercept)
        affected_indices = []
        main_effect_indices = []
        for enc in self._encoders:
            if enc.encodes(vname):
                affected_indices += list(range(start, start + enc.dim))
                if isinstance(enc, MainEffectEncoder):
                    main_effect_indices = list(range(start, start + enc.dim))
            start += enc.dim
        return affected_indices, main_effect_indices


# ===========================================================================
class VariableMap:
    """
    A mapping between columns in an expanded ("feature") matrix and the set of
    variables in the data frame that produced them.
    """

    def __init__(self, affected_features, main_effect_indices):
        """
        Args:
          affected_features: A dictionary, keyed by variable name.
            variable_map[vname] is a list of column indices in the encoded
            predictor matrix that are affected by variable 'vname' (main
            effects and interactions).
          main_effect_indices: A dictionary, keyed by variable name.
            main_effect_indices[vname] is a list of integers corresponding
            coulumn indices in the encoded predictor matrix that are part of
            main effects associated with vname.
        """
        self._variable_names = list(affected_features.keys())
        self._affected_features = affected_features
        self._main_effects_map = main_effect_indices
        self._interaction_mask = None

    # ---------------------------------------------------------------------------
    @property
    def variable_names(self):
        return self._variable_names

    # ---------------------------------------------------------------------------
    def included_variables(self, included_feature_matrix):
        """
        From a matrix of bools denoting included features (variables in the
        expanded matrix), return the corresponding matrix of included variables
        (in the raw data matrix).

        Args:
          included_feature_matrix: A numpy matrix of bools.  Each row denotes a
          subset of included features.  Each column corresponds to a feature in
          the expanded matrix.

        Returns:
          A pandas data frame, with rows corresponding to rows of
          'included_feature_matrix'.  Columns correspond to variables in the
          original data set.  The data frame contains boolean values, with True
          indicating that the variable was present in at least one main effect
          or interaction.
        """
        ans = np.zeros((included_feature_matrix.shape[0],
                        len(self._variable_names)), dtype=bool)

        for i in range(len(self._variable_names)):
            variable_name = self._variable_names[i]
            variable_map = self._affected_features[variable_name]
            variable_indicators = included_feature_matrix[:, variable_map]
            ans[:, i] = np.any(variable_indicators, axis=1)

        return pd.DataFrame(ans, columns=self._variable_names)

    # ---------------------------------------------------------------------------
    def included_interactions(self, included_feature_matrix):
        """
        Returns a pandas DataFrame indicating which two-factor interactions are
        present in the data.

        Args:
          included_feature_matrix: A numpy boolean matrix.
        """
        if self._interaction_mask is None:
            self._create_interaction_mask()

        if not self._interaction_mask:
            return None

        values = {}
        for terms, index in self._interaction_mask.items():
            key = terms[0] + ":" + terms[1]
            values[key] = np.any(included_feature_matrix[:, index], axis=1)
        return pd.DataFrame(values)

    # ---------------------------------------------------------------------------
    def main_effect_indices(self, name: str = None):
        """
        Returns the indices in the encoded matrix of all the main effect columns
        corresponding to the named variable.

        Args:
          name: The name of a variable appearing in a main effect, or None.

        Returns:
          If name is not None then the return value is a list containing the
          feature matrix indices corresponding to the main effect for 'name.'
            Otherwise the return value contains the indices for all main
            effects.
        """
        if name is not None:
            return self._main_effects_map[name]
        else:
            ans = []
            for key in self._main_effects_map.keys():
                ans = ans + self._main_effects_map[key]
            return ans

    # ---------------------------------------------------------------------------
    def interaction_indices(self, name1: str, name2: str):
        """
        The feature indices for the interaction between two variables.

        Args:
          name1: The name of the first variable in the interaction.
          name2: The name of the second variable in the interaction.

        Returns:
          A sorted list of the column indices for the set of features
          describing the interaction.
        """
        name1_indices = set(self._affected_features[name1])
        name2_indices = set(self._affected_features[name2])
        ans = list(name1_indices & name2_indices)
        ans.sort()
        return ans

    # ---------------------------------------------------------------------------
    def _create_interaction_mask(self):
        """
        Binds self._interaction_mask to a dictionary, keyed by pairs of variable
        names in self._variable_names.  The first entry in the pair always
        comes before the second in self._variable_names.  The dictionary entry
        is a list
        """
        self._interaction_mask = {}
        for i in range(len(self._variable_names)):
            for j in range(i + 1, len(self._variable_names)):
                factor1, factor2 = [self._variable_names[k] for k in [i, j]]
                interaction_columns = set(
                    self._affected_features[factor1]).intersection(
                        set(self._variable_map[factor2])
                    )
                if interaction_columns:
                    elements = list(interaction_columns)
                    elements.sort()
                    self._interaction_mask[(factor1, factor2)] = elements
